# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
from packaging import version
import pandas as pd

# syft relative
from . import categorical  # noqa: 401
from . import categorical_dtype  # noqa: 401
from . import frame  # noqa: 401
from . import series  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..misc.union import UnionGenerator
from ..util import generic_update_ast

LIB_NAME = "pandas"
PACKAGE_SUPPORT = {"lib": LIB_NAME}

LIB_VERSION = version.parse(pd.__version__.split("+")[0])


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [("pandas", pd)]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("pandas.DataFrame", "pandas.DataFrame", pd.DataFrame),
        ("pandas.Series", "pandas.Series", pd.Series),
        ("pandas.CategoricalDtype", "pandas.CategoricalDtype", pd.CategoricalDtype),
        ("pandas.Categorical", "pandas.Categorical", pd.Categorical),
    ]

    methods: TypeList[TypeTuple[str, str]] = [
        ("pandas.json_normalize", "pandas.DataFrame"),
        ("pandas.DataFrame.__getitem__", "pandas.Series"),
        ("pandas.DataFrame.__setitem__", "pandas.Series"),
        ("pandas.DataFrame.__len__", "syft.lib.python.Int"),
        ("pandas.DataFrame.__abs__", "pandas.DataFrame"),
        ("pandas.DataFrame.__add__", "pandas.DataFrame"),
        ("pandas.DataFrame.__and__", "pandas.DataFrame"),
        ("pandas.DataFrame.__eq__", "pandas.DataFrame"),
        ("pandas.DataFrame.__floordiv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ge__", "pandas.DataFrame"),
        ("pandas.DataFrame.__gt__", "pandas.DataFrame"),
        ("pandas.DataFrame.__iadd__", "pandas.DataFrame"),
        ("pandas.DataFrame.__iand__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ifloordiv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__imod__", "pandas.DataFrame"),
        ("pandas.DataFrame.__imul__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ipow__", "pandas.DataFrame"),
        ("pandas.DataFrame.__isub__", "pandas.DataFrame"),
        ("pandas.DataFrame.__le__", "pandas.DataFrame"),
        ("pandas.DataFrame.__lt__", "pandas.DataFrame"),
        ("pandas.DataFrame.__mod__", "pandas.DataFrame"),
        ("pandas.DataFrame.__mul__", "pandas.DataFrame"),
        ("pandas.DataFrame.__ne__", "pandas.DataFrame"),
        ("pandas.DataFrame.__neg__", "pandas.DataFrame"),
        ("pandas.DataFrame.__pos__", "pandas.DataFrame"),
        ("pandas.DataFrame.__pow__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rfloordiv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rmod__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rmul__", "pandas.DataFrame"),
        ("pandas.DataFrame.__round__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rpow__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rsub__", "pandas.DataFrame"),
        ("pandas.DataFrame.__rtruediv__", "pandas.DataFrame"),
        ("pandas.DataFrame.__sub__", "pandas.DataFrame"),
        ("pandas.DataFrame.__truediv__", "pandas.DataFrame"),
        ("pandas.DataFrame.dropna", "pandas.DataFrame"),
        ("pandas.DataFrame.apply", "pandas.DataFrame"),
        ("pandas.DataFrame.loc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
        ("pandas.DataFrame.iloc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
        ("pandas.Series.__getitem__", "pandas.Series"),
        ("pandas.Series.__setitem__", "pandas.Series"),
        ("pandas.Series.__len__", "syft.lib.python.Int"),
        ("pandas.Series.__abs__", "pandas.Series"),
        ("pandas.Series.__add__", "pandas.Series"),
        ("pandas.Series.__and__", "pandas.Series"),
        ("pandas.Series.__divmod__", "pandas.Series"),
        ("pandas.Series.__eq__", "pandas.Series"),
        ("pandas.Series.__floordiv__", "pandas.Series"),
        ("pandas.Series.__ge__", "pandas.Series"),
        ("pandas.Series.__gt__", "pandas.Series"),
        ("pandas.Series.__iadd__", "pandas.Series"),
        ("pandas.Series.__iand__", "pandas.Series"),
        ("pandas.Series.__ifloordiv__", "pandas.Series"),
        ("pandas.Series.__imod__", "pandas.Series"),
        ("pandas.Series.__imul__", "pandas.Series"),
        ("pandas.Series.__ipow__", "pandas.Series"),
        ("pandas.Series.__isub__", "pandas.Series"),
        ("pandas.Series.__le__", "pandas.Series"),
        ("pandas.Series.__lt__", "pandas.Series"),
        ("pandas.Series.__mod__", "pandas.Series"),
        ("pandas.Series.__mul__", "pandas.Series"),
        ("pandas.Series.__ne__", "pandas.Series"),
        ("pandas.Series.__neg__", "pandas.Series"),
        ("pandas.Series.__pos__", "pandas.Series"),
        ("pandas.Series.__pow__", "pandas.Series"),
        ("pandas.Series.__rdivmod__", "pandas.Series"),
        ("pandas.Series.__rfloordiv__", "pandas.Series"),
        ("pandas.Series.__rmod__", "pandas.Series"),
        ("pandas.Series.__rmul__", "pandas.Series"),
        ("pandas.Series.__round__", "pandas.Series"),
        ("pandas.Series.__rpow__", "pandas.Series"),
        ("pandas.Series.__rsub__", "pandas.Series"),
        ("pandas.Series.__rtruediv__", "pandas.Series"),
        ("pandas.Series.__sub__", "pandas.Series"),
        ("pandas.Series.__truediv__", "pandas.Series"),
        ("pandas.Series.add", "pandas.Series"),
        ("pandas.Series.apply", "pandas.Series"),
        ("pandas.Series.sub", "pandas.Series"),
        ("pandas.Series.mul", "pandas.Series"),
        ("pandas.Series.div", "pandas.Series"),
        ("pandas.Series.truediv", "pandas.Series"),
        ("pandas.Series.floordiv", "pandas.Series"),
        ("pandas.Series.mod", "pandas.Series"),
        ("pandas.Series.pow", "pandas.Series"),
        ("pandas.Series.radd", "pandas.Series"),
        ("pandas.Series.rsub", "pandas.Series"),
        ("pandas.Series.rmul", "pandas.Series"),
        ("pandas.Series.rdiv", "pandas.Series"),
        ("pandas.Series.rtruediv", "pandas.Series"),
        ("pandas.Series.rfloordiv", "pandas.Series"),
        ("pandas.Series.rmod", "pandas.Series"),
        ("pandas.Series.rpow", "pandas.Series"),
        ("pandas.Series.loc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
        ("pandas.Series.iloc", UnionGenerator["pandas.DataFrame", "pandas.Series"]),
        ("pandas.Series.lt", "pandas.Series"),
        ("pandas.Series.gt", "pandas.Series"),
        ("pandas.Series.le", "pandas.Series"),
        ("pandas.Series.ge", "pandas.Series"),
        ("pandas.Series.ne", "pandas.Series"),
        ("pandas.Series.eq", "pandas.Series"),
        ("pandas.Series.argsort", "pandas.Series"),
        ("pandas.Series.round", "pandas.Series"),
        ("pandas.Series.head", "pandas.Series"),
        ("pandas.Series.tail", "pandas.Series"),
        ("pandas.Series.any", "syft.lib.python.Bool"),
        ("pandas.Series.shape", "syft.lib.python.Tuple"),
        ("pandas.Series.all", "syft.lib.python.Bool"),
        ("pandas.Series.argmax", "syft.lib.python.Int"),
        ("pandas.Series.nbytes", "syft.lib.python.Int"),
        ("pandas.Series.mean", "syft.lib.python.Float"),
        ("pandas.Series.ndim", "syft.lib.python.Int"),
        ("pandas.Series.size", "syft.lib.python.Int"),
        ("pandas.Series.hasnans", "syft.lib.python.Bool"),
        ("pandas.Series.empty", "syft.lib.python.Bool"),
        ("pandas.Series.T", "pandas.Series"),
        ("pandas.Series.dropna", "pandas.Series"),
        ("pandas.Series.to_frame", "pandas.DataFrame"),
        ("pandas.Series.to_list", "syft.lib.python.List"),
        (
            "pandas.Series.sum",
            UnionGenerator["syft.lib.python.Float", "syft.lib.python.Int"],
        ),
        ("pandas.Series.median", "syft.lib.python.Float"),
        (
            "pandas.Series.max",
            UnionGenerator[
                "syft.lib.python.Bool", "syft.lib.python.Float", "syft.lib.python.Int"
            ],
        ),
        (
            "pandas.Series.min",
            UnionGenerator[
                "syft.lib.python.Bool", "syft.lib.python.Float", "syft.lib.python.Int"
            ],
        ),
        # ===== methods for pd.Categorical ======
        # ("pandas.Categorical.__array__",),
        # ("pandas.Categorical.__array_ufunc__",),
        # ("pandas.Categorical.__contains__",),
        # ("pandas.Categorical.__dir__",),
        # ("pandas.Categorical.__init__",),
        # ("pandas.Categorical.__iter__",), Return a list Iterator
        # ("pandas.Categorical.__repr__",),
        ("pandas.Categorical.__eq__", "numpy.ndarray"),
        ("pandas.Categorical.__ge__", "numpy.ndarray"),
        ("pandas.Categorical.__gt__", "numpy.ndarray"),
        ("pandas.Categorical.__le__", "numpy.ndarray"),
        ("pandas.Categorical.__lt__", "numpy.ndarray"),
        ("pandas.Categorical.__ne__", "numpy.ndarray"),
        ("pandas.Categorical.__len__", "syft.lib.python.Int"),
        (
            "pandas.Categorical.__getitem__",
            UnionGenerator[
                "syft.lib.python.Bool",
                "syft.lib.python.Float",
                "syft.lib.python.Int",
                "syft.lib.python.Complex",
            ],
        ),
        ("pandas.Categorical.__setitem__", "syft.lib.python._SyNone"),
        # ("pandas.Categorical.__setstate__",), # method to support Pickle support
        # ("pandas.Categorical.__sizeof__",),
        ("pandas.Categorical.add_categories", "pandas.Categorical"),
        ("pandas.Categorical.argmax", "syft.lib.python.Int"),
        ("pandas.Categorical.argmin", "syft.lib.python.Int"),
        ("pandas.Categorical.argsort", "numpy.ndarray"),
        ("pandas.Categorical.as_ordered", "pandas.Categorical"),
        ("pandas.Categorical.as_unordered", "pandas.Categorical"),
        # ("pandas.Categorical.astype",), # Need support for other pandas types
        # ("pandas.Categorical.check_for_ordered",),
        ("pandas.Categorical.copy", "pandas.Categorical"),
        ("pandas.Categorical.describe", "pandas.DataFrame"),
        ("pandas.Categorical.dropna", "pandas.Categorical"),
        ("pandas.Categorical.equals", "syft.lib.python.Bool"),
        (
            "pandas.Categorical.factorize",
            "syft.lib.python.Tuple",
        ),
        ("pandas.Categorical.fillna", "pandas.Categorical"),
        ("pandas.Categorical.from_codes", "pandas.Categorical"),
        ("pandas.Categorical.is_dtype_equal", "syft.lib.python.Bool"),
        ("pandas.Categorical.isin", "numpy.ndarray"),
        ("pandas.Categorical.isna", "numpy.ndarray"),
        ("pandas.Categorical.isnull", "numpy.ndarray"),
        (
            "pandas.Categorical.map",
            "pandas.Categorical",
        ),  # TODO: returns pd.Categorical or pd.Index
        (
            "pandas.Categorical.max",
            UnionGenerator[
                "syft.lib.python.Bool",
                "syft.lib.python.Float",
                "syft.lib.python.Int",
                "syft.lib.python.Complex",
            ],
        ),
        ("pandas.Categorical.memory_usage", "syft.lib.python.Int"),
        (
            "pandas.Categorical.min",
            UnionGenerator[
                "syft.lib.python.Bool",
                "syft.lib.python.Float",
                "syft.lib.python.Int",
                "syft.lib.python.Complex",
            ],
        ),
        ("pandas.Categorical.mode", "pandas.Categorical"),
        ("pandas.Categorical.notna", "numpy.ndarray"),
        ("pandas.Categorical.notnull", "numpy.ndarray"),
        ("pandas.Categorical.ravel", "pandas.Categorical"),
        ("pandas.Categorical.remove_categories", "pandas.Categorical"),
        ("pandas.Categorical.remove_unused_categories", "pandas.Categorical"),
        (
            "pandas.Categorical.rename_categories",
            "pandas.Categorical",
        ),
        (
            "pandas.Categorical.reorder_categories",
            "pandas.Categorical",
        ),
        ("pandas.Categorical.repeat", "pandas.Categorical"),
        ("pandas.Categorical.replace", "pandas.Categorical"),
        ("pandas.Categorical.reshape", "pandas.Categorical"),
        ("pandas.Categorical.searchsorted", "syft.lib.python.Int"),
        (
            "pandas.Categorical.set_categories",
            "pandas.Categorical",
        ),
        (
            "pandas.Categorical.set_ordered",
            "pandas.Categorical",
        ),
        ("pandas.Categorical.shift", "pandas.Categorical"),
        ("pandas.Categorical.sort_values", "pandas.Categorical"),
        ("pandas.Categorical.take", "pandas.Categorical"),
        ("pandas.Categorical.take_nd", "pandas.Categorical"),
        ("pandas.Categorical.to_dense", "numpy.ndarray"),
        ("pandas.Categorical.to_list", "syft.lib.python.Float"),
        ("pandas.Categorical.to_numpy", "numpy.ndarray"),
        ("pandas.Categorical.tolist", "syft.lib.python.Float"),
        ("pandas.Categorical.unique", "pandas.Categorical"),
        ("pandas.Categorical.value_counts", "pandas.Series"),
        ("pandas.Categorical.view", "pandas.Categorical"),
        # ==== pd.Categorical properties ====
        ("pandas.Categorical.T", "pandas.Categorical"),
        # ("pandas.Categorical.categories",), require support for pd.Index
        ("pandas.Categorical.codes", "numpy.ndarray"),
        ("pandas.Categorical.dtype", "pandas.CategoricalDtype"),
        ("pandas.Categorical.nbytes", "syft.lib.python.Int"),
        ("pandas.Categorical.ordered", "syft.lib.python.Bool"),
        ("pandas.Categorical.shape", "syft.lib.python.Tuple"),
        ("pandas.Categorical.ndim", "syft.lib.python.Int"),
    ]

    if LIB_VERSION > version.parse("1.2.0"):
        methods += [
            ("pandas.DataFrame.__divmod__", "pandas.DataFrame"),
            ("pandas.DataFrame.__rdivmod__", "pandas.DataFrame"),
        ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        # klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



"""Serde method for LogisticRegression (currently)!"""

# third party
import sklearn  # noqa: 401

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.python.dict import Dict
from ...lib.python.primitive_factory import PrimitiveFactory
from ...proto.lib.sklearn.logistic_model_pb2 import Logistic as Logistic_PB


def object2proto(obj: sklearn.linear_model.LogisticRegression) -> Logistic_PB:
    """Object to proto conversion using Logistic_PB.

    Args:
        obj: Model which we have to send to client

    Returns:
        Serialized proto.

    """
    vars_dict = PrimitiveFactory.generate_primitive(value=vars(obj))
    dict_proto = vars_dict._object2proto()

    return Logistic_PB(
        id=dict_proto.id,
        model=dict_proto,
    )


def proto2object(proto: Logistic_PB) -> sklearn.linear_model.LogisticRegression:
    """Proto to object conversion using to return desired model.

    Args:
        proto: Serialized version of model, which will be used to re-construct model.

    Returns:
        Re-constrcuted model.
    """
    vars_dict = Dict._proto2object(proto=proto.model)
    ret_model = sklearn.linear_model.LogisticRegression()

    for attribute, value in vars_dict.items():
        if hasattr(value, "upcast"):
            # This would convert all Sy objects to their original types
            value = value.upcast()
        setattr(ret_model, attribute, value)

    return ret_model


GenerateWrapper(
    wrapped_type=sklearn.linear_model.LogisticRegression,
    import_path="sklearn.linear_model.LogisticRegression",
    protobuf_scheme=Logistic_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
from typing import Dict
from typing import Union

# (path: str, return_type:type)
allowlist: Dict[str, Union[str, Dict[str, str]]] = {}

allowlist["torchvision.__version__"] = "syft.lib.python.String"
# MNIST
allowlist["torchvision.transforms.Compose"] = "torchvision.transforms.Compose"
# allowlist["torchvision.transforms.Compose.__iter__"] = "torchvision.transforms.ToTensor"
# TODO: Compose.transforms property only exists on the object not on the class?
# allowlist["torchvision.transforms.Compose.transforms"] = "syft.lib.python.List"
allowlist["torchvision.transforms.ToTensor"] = "torchvision.transforms.ToTensor"
allowlist["torchvision.transforms.Normalize"] = "torchvision.transforms.Normalize"
# TODO: Normalize properties only exists on the object not on the class?
# allowlist["torchvision.transforms.Normalize.inplace"] = "syft.lib.python.Bool"
# TODO: mean and std are actually tuples
# allowlist["torchvision.transforms.Normalize.mean"] = "syft.lib.python.List"
# allowlist["torchvision.transforms.Normalize.std"] = "syft.lib.python.List"

# Datasets

allowlist["torchvision.datasets.MNIST"] = "torchvision.datasets.MNIST"

allowlist["torchvision.datasets.MNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CelebA"] = "torchvision.datasets.CelebA"
allowlist["torchvision.datasets.CelebA.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CIFAR10"] = "torchvision.datasets.CIFAR10"
allowlist["torchvision.datasets.CIFAR10.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CIFAR100"] = "torchvision.datasets.CIFAR100"
allowlist["torchvision.datasets.CIFAR10.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Cityscapes"] = "torchvision.datasets.Cityscapes"
allowlist["torchvision.datasets.Cityscapes.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CocoCaptions"] = "torchvision.datasets.CocoCaptions"
allowlist["torchvision.datasets.CocoCaptions.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.CocoDetection"] = "torchvision.datasets.CocoDetection"
allowlist["torchvision.datasets.CocoDetection.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.EMNIST"] = "torchvision.datasets.EMNIST"
allowlist["torchvision.datasets.EMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.FakeData"] = "torchvision.datasets.FakeData"
allowlist["torchvision.datasets.FakeData.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.FashionMNIST"] = "torchvision.datasets.FashionMNIST"
allowlist["torchvision.datasets.FashionMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Flickr8k"] = "torchvision.datasets.Flickr8k"
allowlist["torchvision.datasets.Flickr8k.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Flickr30k"] = "torchvision.datasets.Flickr30k"
allowlist["torchvision.datasets.Flickr30k.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.HMDB51"] = "torchvision.datasets.HMDB51"
allowlist["torchvision.datasets.HMDB51.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.ImageNet"] = "torchvision.datasets.ImageNet"
allowlist["torchvision.datasets.ImageNet.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Kinetics400"] = "torchvision.datasets.Kinetics400"
allowlist["torchvision.datasets.Kinetics400.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.KMNIST"] = "torchvision.datasets.KMNIST"
allowlist["torchvision.datasets.KMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.LSUN"] = "torchvision.datasets.LSUN"
allowlist["torchvision.datasets.LSUN.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Omniglot"] = {
    "return_type": "torchvision.datasets.Omniglot",
    "min_version": "0.8.0",
}
allowlist["torchvision.datasets.Omniglot.__len__"] = {
    "return_type": "syft.lib.python.Int",
    "min_version": "0.8.0",
}

allowlist["torchvision.datasets.PhotoTour"] = "torchvision.datasets.PhotoTour"
allowlist["torchvision.datasets.PhotoTour.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.Places365"] = {
    "return_type": "torchvision.datasets.Places365",
    "min_version": "0.8.0",
}
allowlist["torchvision.datasets.Places365.__len__"] = {
    "return_type": "syft.lib.python.Int",
    "min_version": "0.8.0",
}

allowlist["torchvision.datasets.QMNIST"] = "torchvision.datasets.QMNIST"
allowlist["torchvision.datasets.QMNIST.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.SBDataset"] = "torchvision.datasets.SBDataset"
allowlist["torchvision.datasets.SBDataset.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.SBU"] = "torchvision.datasets.SBU"
allowlist["torchvision.datasets.SBU.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.STL10"] = "torchvision.datasets.STL10"
allowlist["torchvision.datasets.STL10.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.SVHN"] = "torchvision.datasets.SVHN"
allowlist["torchvision.datasets.SVHN.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.UCF101"] = "torchvision.datasets.UCF101"
allowlist["torchvision.datasets.UCF101.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.USPS"] = "torchvision.datasets.USPS"
allowlist["torchvision.datasets.USPS.__len__"] = "syft.lib.python.Int"

allowlist[
    "torchvision.datasets.VOCSegmentation"
] = "torchvision.datasets.VOCSegmentation"
allowlist["torchvision.datasets.VOCSegmentation.__len__"] = "syft.lib.python.Int"

allowlist["torchvision.datasets.VOCDetection"] = "torchvision.datasets.VOCDetection"
allowlist[
    "torchvision.datasets.VOCDetection.__len__"
] = "torchvision.datasets.VOCDetection"

# Transforms
allowlist["torchvision.transforms.CenterCrop"] = "torchvision.transforms.CenterCrop"
allowlist["torchvision.transforms.ColorJitter"] = "torchvision.transforms.ColorJitter"


allowlist["torchvision.transforms.FiveCrop"] = "torchvision.transforms.FiveCrop"


allowlist["torchvision.transforms.Grayscale"] = "torchvision.transforms.Grayscale"
allowlist["torchvision.transforms.Pad"] = "torchvision.transforms.Pad"
allowlist["torchvision.transforms.RandomAffine"] = "torchvision.transforms.RandomAffine"

allowlist["torchvision.transforms.RandomApply"] = "torchvision.transforms.RandomApply"

allowlist["torchvision.transforms.RandomCrop"] = "torchvision.transforms.RandomCrop"
allowlist[
    "torchvision.transforms.RandomGrayscale"
] = "torchvision.transforms.RandomGrayscale"

allowlist[
    "torchvision.transforms.RandomHorizontalFlip"
] = "torchvision.transforms.RandomHorizontalFlip"
allowlist[
    "torchvision.transforms.RandomPerspective"
] = "torchvision.transforms.RandomPerspective"

allowlist[
    "torchvision.transforms.RandomResizedCrop"
] = "torchvision.transforms.RandomResizedCrop"
allowlist[
    "torchvision.transforms.RandomRotation"
] = "torchvision.transforms.RandomRotation"
allowlist[
    "torchvision.transforms.RandomSizedCrop"
] = "torchvision.transforms.RandomSizedCrop"
allowlist[
    "torchvision.transforms.RandomVerticalFlip"
] = "torchvision.transforms.RandomVerticalFlip"
allowlist["torchvision.transforms.Resize"] = "torchvision.transforms.Resize"
allowlist["torchvision.transforms.Scale"] = "torchvision.transforms.Scale"
allowlist["torchvision.transforms.TenCrop"] = "torchvision.transforms.TenCrop"
allowlist["torchvision.transforms.GaussianBlur"] = {
    "return_type": "torchvision.transforms.GaussianBlur",
    "min_version": "0.8.0",
}

allowlist["torchvision.transforms.RandomChoice"] = "torchvision.transforms.RandomChoice"
allowlist["torchvision.transforms.RandomOrder"] = "torchvision.transforms.RandomOrder"

allowlist[
    "torchvision.transforms.LinearTransformation"
] = "torchvision.transforms.LinearTransformation"
allowlist[
    "torchvision.transforms.RandomErasing"
] = "torchvision.transforms.RandomErasing"
allowlist["torchvision.transforms.ConvertImageDtype"] = {
    "return_type": "torchvision.transforms.ConvertImageDtype",
    "min_version": "0.8.0",
}
allowlist["torchvision.transforms.ToPILImage"] = "torchvision.transforms.ToPILImage"
allowlist["torchvision.transforms.Lambda"] = "torchvision.transforms.Lambda"

# Functional Transformers
allowlist["torchvision.transforms.functional.adjust_brightness"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.adjust_contrast"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.adjust_gamma"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.adjust_hue"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.adjust_saturation"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.adjust_sharpness"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
}
allowlist["torchvision.transforms.functional.affine"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.autocontrast"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
}
allowlist["torchvision.transforms.functional.center_crop"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.convert_image_dtype"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.7.0",
}
allowlist["torchvision.transforms.functional.crop"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.equalize"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
}
allowlist["torchvision.transforms.functional.erase"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.five_crop"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.gaussian_blur"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
}
allowlist["torchvision.transforms.functional.hflip"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.invert"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
}
allowlist["torchvision.transforms.functional.normalize"] = "torch.Tensor"
# "test_parameters (tests/lib/torchvision/allowlist_test_parameters.json): (tens, [0.5,
#  0.5, 0.5], [1, 1, 1]).unsqueeze(0)",
# currently not added because of test issues with hier versions
# (//) works for 1.6.0 and / works for higher version :(

allowlist["torchvision.transforms.functional.pad"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.6.0",
}
allowlist["torchvision.transforms.functional.perspective"] = "torch.Tensor"

allowlist["torchvision.transforms.functional.pil_to_tensor"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.posterize"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
}
allowlist["torchvision.transforms.functional.resize"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.resized_crop"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.rgb_to_grayscale"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.8.0",
}
allowlist["torchvision.transforms.functional.rotate"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.solarize"] = {
    "return_type": "torch.Tensor",
    "min_version": "0.9.0",
}
allowlist["torchvision.transforms.functional.ten_crop"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.to_grayscale"] = "PIL.Image.Image"
allowlist["torchvision.transforms.functional.to_pil_image"] = "PIL.Image.Image"
allowlist["torchvision.transforms.functional.to_tensor"] = "torch.Tensor"
allowlist["torchvision.transforms.functional.vflip"] = "torch.Tensor"



# third party
import transformers
from transformers import PretrainedConfig

# syft relative
from ... import deserialize
from ...generate_wrapper import GenerateWrapper
from ...lib.python.primitive_factory import PrimitiveFactory
from ...lib.python.util import upcast
from ...proto.lib.transformers.hf_config_pb2 import HFConfig as HFConfig_PB


def object2proto(obj: PretrainedConfig) -> HFConfig_PB:
    config_class = PrimitiveFactory.generate_primitive(value=type(obj).__name__)
    config_attrs = PrimitiveFactory.generate_primitive(value=obj.to_dict())

    config_proto = HFConfig_PB(
        id=config_class.id._object2proto(),
        config_class=config_class._object2proto(),
        config_attrs=config_attrs._object2proto(),
    )
    return config_proto


def proto2object(proto: HFConfig_PB) -> PretrainedConfig:
    config_class = getattr(transformers, upcast(deserialize(proto.config_class)))
    config = config_class.from_dict(upcast(deserialize(proto.config_attrs)))
    return config


GenerateWrapper(
    wrapped_type=PretrainedConfig,
    import_path="transformers.PretrainedConfig",
    protobuf_scheme=HFConfig_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
import dataclasses
from uuid import UUID

# third party
import sympc
from sympc.config import Config
from sympc.tensor import ReplicatedSharedTensor

# syft absolute
import syft

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...proto.lib.sympc.replicatedshared_tensor_pb2 import (
    ReplicatedSharedTensor as ReplicatedSharedTensor_PB,
)
from ..python.primitive_factory import PrimitiveFactory


def object2proto(obj: object) -> ReplicatedSharedTensor_PB:
    share: ReplicatedSharedTensor = obj

    session_uuid = ""
    config = {}

    if share.session_uuid is not None:
        session_uuid = str(share.session_uuid)

    config = dataclasses.asdict(share.config)
    session_uuid_syft = session_uuid
    conf_syft = syft.serialize(
        PrimitiveFactory.generate_primitive(value=config), to_proto=True
    )
    proto = ReplicatedSharedTensor_PB(session_uuid=session_uuid_syft, config=conf_syft)

    for tensor in share.shares:
        proto.tensor.append(syft.serialize(tensor, to_proto=True))

    return proto


def proto2object(proto: ReplicatedSharedTensor_PB) -> ReplicatedSharedTensor:
    if proto.session_uuid:
        session = sympc.session.get_session(proto.session_uuid)
        if session is None:
            raise ValueError(f"The session {proto.session_uuid} could not be found")

        config = dataclasses.asdict(session.config)
    else:
        config = syft.deserialize(proto.config, from_proto=True)

    output_shares = []

    for tensor in proto.tensor:
        output_shares.append(syft.deserialize(tensor, from_proto=True))

    share = ReplicatedSharedTensor(shares=None, config=Config(**config))

    if proto.session_uuid:
        share.session_uuid = UUID(proto.session_uuid)

    share.shares = output_shares

    return share


GenerateWrapper(
    wrapped_type=ReplicatedSharedTensor,
    import_path="sympc.tensor.ReplicatedSharedTensor",
    protobuf_scheme=ReplicatedSharedTensor_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib

# stdlib
import dataclasses
from uuid import UUID

# third party
import sympc
from sympc.config import Config
from sympc.tensor import ShareTensor

# syft absolute
import syft

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...proto.lib.sympc.share_tensor_pb2 import ShareTensor as ShareTensor_PB
from ..python.primitive_factory import PrimitiveFactory


def object2proto(obj: object) -> ShareTensor_PB:
    share: ShareTensor = obj

    session_uuid = ""
    config = {}

    if share.session_uuid is not None:
        session_uuid = str(share.session_uuid)

    config = dataclasses.asdict(share.config)
    session_uuid_syft = session_uuid
    conf_syft = syft.serialize(
        PrimitiveFactory.generate_primitive(value=config), to_proto=True
    )
    proto = ShareTensor_PB(session_uuid=session_uuid_syft, config=conf_syft)

    tensor_data = getattr(share.tensor, "data", None)
    if tensor_data is not None:
        proto.tensor.CopyFrom(syft.serialize(share.tensor, to_proto=True))

    return proto


def proto2object(proto: ShareTensor_PB) -> ShareTensor:
    if proto.session_uuid:
        session = sympc.session.get_session(proto.session_uuid)
        if session is None:
            raise ValueError(f"The session {proto.session_uuid} could not be found")

        config = dataclasses.asdict(session.config)
    else:
        config = syft.deserialize(proto.config, from_proto=True)

    tensor = syft.deserialize(proto.tensor, from_proto=True)
    share = ShareTensor(data=None, config=Config(**config))

    if proto.session_uuid:
        share.session_uuid = UUID(proto.session_uuid)

    # Manually put the tensor since we do not want to re-encode it
    share.tensor = tensor

    return share


GenerateWrapper(
    wrapped_type=ShareTensor,
    import_path="sympc.tensor.ShareTensor",
    protobuf_scheme=ShareTensor_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
from typing import Dict
from typing import Union

allowlist: Dict[str, Union[str, Dict[str, str]]] = {}  # (path: str, return_type:type)


# General
allowlist[
    "transformers.tokenization_utils_base.BatchEncoding"
] = "transformers.tokenization_utils_base.BatchEncoding"
allowlist[
    "transformers.tokenization_utils_base.BatchEncoding.__getitem__"
] = "torch.Tensor"
allowlist[
    "transformers.tokenization_utils_fast.PreTrainedTokenizerFast"
] = "transformers.tokenization_utils_fast.PreTrainedTokenizerFast"
allowlist[
    "transformers.tokenization_utils_fast.PreTrainedTokenizerFast.__call__"
] = "transformers.tokenization_utils_base.BatchEncoding"


# Distilbert model
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel"
] = "transformers.models.distilbert.modeling_distilbert.DistilBertModel"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.__call__"
] = "torch.Tensor"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.parameters"
] = "syft.lib.python.List"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.register_parameter"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.register_parameter"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.train"
] = "transformers.models.distilbert.modeling_distilbert.DistilBertModel"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.eval"
] = "transformers.models.distilbert.modeling_distilbert.DistilBertModel"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.cuda"
] = "transformers.models.distilbert.modeling_distilbert.DistilBertModel"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.cpu"
] = "transformers.models.distilbert.modeling_distilbert.DistilBertModel"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.state_dict"
] = "syft.lib.python.collections.OrderedDict"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.load_state_dict"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.DistilBertModel.extra_repr"
] = "syft.lib.python.String"

# DistilBert modules
allowlist[
    "transformers.models.distilbert.modeling_distilbert.Embeddings"
] = "transformers.models.distilbert.modeling_distilbert.Embeddings"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.MultiHeadSelfAttention"
] = "transformers.models.distilbert.modeling_distilbert.MultiHeadSelfAttention"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.FFN"
] = "transformers.models.distilbert.modeling_distilbert.FFN"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.TransformerBlock"
] = "transformers.models.distilbert.modeling_distilbert.TransformerBlock"
allowlist[
    "transformers.models.distilbert.modeling_distilbert.Transformer"
] = "transformers.models.distilbert.modeling_distilbert.Transformer"

# DistilBert misc
allowlist[
    "transformers.models.distilbert.configuration_distilbert.DistilBertConfig"
] = "transformers.models.distilbert.configuration_distilbert.DistilBertConfig"
allowlist[
    "transformers.models.distilbert.tokenization_distilbert_fast.DistilBertTokenizerFast"
] = "transformers.models.distilbert.tokenization_distilbert_fast.DistilBertTokenizerFast"
allowlist[
    "transformers.models.distilbert.tokenization_distilbert_fast.DistilBertTokenizerFast.__call__"
] = "transformers.tokenization_utils_base.BatchEncoding"

# XLMRoberta
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel"
] = "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.from_pretrained"
] = "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.__call__"
] = "torch.Tensor"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.parameters"
] = "syft.lib.python.List"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.register_parameter"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.register_parameter"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.train"
] = "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.eval"
] = "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.cuda"
] = "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.cpu"
] = "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.state_dict"
] = "syft.lib.python.collections.OrderedDict"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.load_state_dict"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.xlm_roberta.modeling_xlm_roberta.XLMRobertaModel.extra_repr"
] = "syft.lib.python.String"

# Roberta classification
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"
] = "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.__call__"
] = "torch.Tensor"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.parameters"
] = "syft.lib.python.List"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.register_parameter"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.register_parameter"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.train"
] = "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.eval"
] = "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.cuda"
] = "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.cpu"
] = "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.state_dict"
] = "syft.lib.python.collections.OrderedDict"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.load_state_dict"
] = "syft.lib.python._SyNone"
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead.extra_repr"
] = "syft.lib.python.String"


# Roberta submodules
allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaEmbeddings"
] = "transformers.models.roberta.modeling_roberta.RobertaEmbeddings"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaSelfAttention"
] = "transformers.models.roberta.modeling_roberta.RobertaSelfAttention"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaSelfOutput"
] = "transformers.models.roberta.modeling_roberta.RobertaSelfOutput"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaAttention"
] = "transformers.models.roberta.modeling_roberta.RobertaAttention"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaIntermediate"
] = "transformers.models.roberta.modeling_roberta.RobertaIntermediate"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaOutput"
] = "transformers.models.roberta.modeling_roberta.RobertaOutput"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaLayer"
] = "transformers.models.roberta.modeling_roberta.RobertaLayer"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaEncoder"
] = "transformers.models.roberta.modeling_roberta.RobertaEncoder"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaPooler"
] = "transformers.models.roberta.modeling_roberta.RobertaPooler"

allowlist[
    "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"
] = "transformers.models.roberta.modeling_roberta.RobertaClassificationHead"

# XLMRoberta misc

allowlist[
    "transformers.models.xlm_roberta.configuration_xlm_roberta.XLMRobertaConfig"
] = "transformers.models.xlm_roberta.configuration_xlm_roberta.XLMRobertaConfig"

allowlist[
    "transformers.models.xlm_roberta.tokenization_xlm_roberta_fast.XLMRobertaTokenizerFast"
] = "transformers.models.xlm_roberta.tokenization_xlm_roberta_fast.XLMRobertaTokenizerFast"

allowlist[
    "transformers.models.xlm_roberta.tokenization_xlm_roberta_fast.XLMRobertaTokenizerFast.from_pretrained"
] = "transformers.models.xlm_roberta.tokenization_xlm_roberta_fast.XLMRobertaTokenizerFast"

allowlist[
    "transformers.models.xlm_roberta.tokenization_xlm_roberta_fast.XLMRobertaTokenizerFast.__call__"
] = "transformers.tokenization_utils_base.BatchEncoding"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/core/store/dataset.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.store import (
    store_object_pb2 as proto_dot_core_dot_store_dot_store__object__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/core/store/dataset.proto",
    package="syft.core.store",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1eproto/core/store/dataset.proto\x12\x0fsyft.core.store\x1a%proto/core/common/common_object.proto\x1a#proto/core/store/store_object.proto"\xe2\x01\n\x07\x44\x61taset\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x10\n\x08obj_type\x18\x02 \x01(\t\x12\x1a\n\x12schematic_qualname\x18\x03 \x01(\t\x12-\n\x04\x64\x61ta\x18\x04 \x03(\x0b\x32\x1f.syft.core.store.StorableObject\x12\x13\n\x0b\x64\x65scription\x18\x05 \x01(\t\x12\x0c\n\x04tags\x18\x06 \x03(\t\x12\x18\n\x10read_permissions\x18\x07 \x01(\x0c\x12\x1a\n\x12search_permissions\x18\x08 \x01(\x0c\x62\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_core_dot_store_dot_store__object__pb2.DESCRIPTOR,
    ],
)


_DATASET = _descriptor.Descriptor(
    name="Dataset",
    full_name="syft.core.store.Dataset",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="syft.core.store.Dataset.id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="obj_type",
            full_name="syft.core.store.Dataset.obj_type",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="schematic_qualname",
            full_name="syft.core.store.Dataset.schematic_qualname",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="syft.core.store.Dataset.data",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="description",
            full_name="syft.core.store.Dataset.description",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="tags",
            full_name="syft.core.store.Dataset.tags",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="read_permissions",
            full_name="syft.core.store.Dataset.read_permissions",
            index=6,
            number=7,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="search_permissions",
            full_name="syft.core.store.Dataset.search_permissions",
            index=7,
            number=8,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=128,
    serialized_end=354,
)

_DATASET.fields_by_name[
    "id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_DATASET.fields_by_name[
    "data"
].message_type = proto_dot_core_dot_store_dot_store__object__pb2._STORABLEOBJECT
DESCRIPTOR.message_types_by_name["Dataset"] = _DATASET
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Dataset = _reflection.GeneratedProtocolMessageType(
    "Dataset",
    (_message.Message,),
    {
        "DESCRIPTOR": _DATASET,
        "__module__": "proto.core.store.dataset_pb2"
        # @@protoc_insertion_point(class_scope:syft.core.store.Dataset)
    },
)
_sym_db.RegisterMessage(Dataset)


# @@protoc_insertion_point(module_scope)



# stdlib
from enum import Enum


class PyGridClientEnums(str, Enum):
    ENCODING = "ISO-8859-1"


class RequestAPIFields(str, Enum):
    ADDRESS = "address"
    UID = "uid"
    POINTABLE = "pointable"
    QUERY = "query"
    CONTENT = "content"
    REPLY_TO = "reply_to"
    MESSAGE = "message"
    ERROR = "error"
    VALUE = "value"
    HANDSHAKE = "handshake"
    SENDER_ADDRESS = "sender_address"
    DOMAIN_ADDRESS = "domain_address"


class AssociationRequestResponses(str, Enum):
    ACCEPT = "accept"
    DENY = "deny"


class ResponseObjectEnum(str, Enum):
    ASSOCIATION_REQUEST = "association-request"
    DATASET = "dataset"
    GROUP = "group"
    ROLE = "role"
    USER = "user"
    WORKER = "worker"
    DATA = "data"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/parameter.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/torch/parameter.proto",
    package="syft.lib.torch",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1fproto/lib/torch/parameter.proto\x12\x0esyft.lib.torch"}\n\x0eParameterProto\x12\x0e\n\x06tensor\x18\x01 \x01(\x0c\x12\x15\n\rrequires_grad\x18\x02 \x01(\x08\x12\x11\n\x04grad\x18\x03 \x01(\x0cH\x00\x88\x01\x01\x12\x18\n\x0bgrad_sample\x18\x04 \x01(\x0cH\x01\x88\x01\x01\x42\x07\n\x05_gradB\x0e\n\x0c_grad_sampleb\x06proto3',
)


_PARAMETERPROTO = _descriptor.Descriptor(
    name="ParameterProto",
    full_name="syft.lib.torch.ParameterProto",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="tensor",
            full_name="syft.lib.torch.ParameterProto.tensor",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="requires_grad",
            full_name="syft.lib.torch.ParameterProto.requires_grad",
            index=1,
            number=2,
            type=8,
            cpp_type=7,
            label=1,
            has_default_value=False,
            default_value=False,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="grad",
            full_name="syft.lib.torch.ParameterProto.grad",
            index=2,
            number=3,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="grad_sample",
            full_name="syft.lib.torch.ParameterProto.grad_sample",
            index=3,
            number=4,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"",
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_grad",
            full_name="syft.lib.torch.ParameterProto._grad",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_grad_sample",
            full_name="syft.lib.torch.ParameterProto._grad_sample",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=51,
    serialized_end=176,
)

_PARAMETERPROTO.oneofs_by_name["_grad"].fields.append(
    _PARAMETERPROTO.fields_by_name["grad"]
)
_PARAMETERPROTO.fields_by_name[
    "grad"
].containing_oneof = _PARAMETERPROTO.oneofs_by_name["_grad"]
_PARAMETERPROTO.oneofs_by_name["_grad_sample"].fields.append(
    _PARAMETERPROTO.fields_by_name["grad_sample"]
)
_PARAMETERPROTO.fields_by_name[
    "grad_sample"
].containing_oneof = _PARAMETERPROTO.oneofs_by_name["_grad_sample"]
DESCRIPTOR.message_types_by_name["ParameterProto"] = _PARAMETERPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ParameterProto = _reflection.GeneratedProtocolMessageType(
    "ParameterProto",
    (_message.Message,),
    {
        "DESCRIPTOR": _PARAMETERPROTO,
        "__module__": "proto.lib.torch.parameter_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.ParameterProto)
    },
)
_sym_db.RegisterMessage(ParameterProto)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/transformers/tokenizerfast.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/transformers/tokenizerfast.proto",
    package="syft.lib.transformers",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n*proto/lib/transformers/tokenizerfast.proto\x12\x15syft.lib.transformers\x1a%proto/core/common/common_object.proto\x1a\x1bproto/lib/python/dict.proto"\x84\x01\n\rTokenizerFast\x12!\n\x02id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x16\n\x0etokenizer_type\x18\x02 \x01(\t\x12\x11\n\ttokenizer\x18\x03 \x01(\t\x12%\n\x06kwargs\x18\x04 \x01(\x0b\x32\x15.syft.lib.python.Dictb\x06proto3',
    dependencies=[
        proto_dot_core_dot_common_dot_common__object__pb2.DESCRIPTOR,
        proto_dot_lib_dot_python_dot_dict__pb2.DESCRIPTOR,
    ],
)


_TOKENIZERFAST = _descriptor.Descriptor(
    name="TokenizerFast",
    full_name="syft.lib.transformers.TokenizerFast",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="id",
            full_name="syft.lib.transformers.TokenizerFast.id",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="tokenizer_type",
            full_name="syft.lib.transformers.TokenizerFast.tokenizer_type",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="tokenizer",
            full_name="syft.lib.transformers.TokenizerFast.tokenizer",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="kwargs",
            full_name="syft.lib.transformers.TokenizerFast.kwargs",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=138,
    serialized_end=270,
)

_TOKENIZERFAST.fields_by_name[
    "id"
].message_type = proto_dot_core_dot_common_dot_common__object__pb2._UID
_TOKENIZERFAST.fields_by_name[
    "kwargs"
].message_type = proto_dot_lib_dot_python_dot_dict__pb2._DICT
DESCRIPTOR.message_types_by_name["TokenizerFast"] = _TOKENIZERFAST
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TokenizerFast = _reflection.GeneratedProtocolMessageType(
    "TokenizerFast",
    (_message.Message,),
    {
        "DESCRIPTOR": _TOKENIZERFAST,
        "__module__": "proto.lib.transformers.tokenizerfast_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.transformers.TokenizerFast)
    },
)
_sym_db.RegisterMessage(TokenizerFast)


# @@protoc_insertion_point(module_scope)



# stdlib
import inspect
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional

# syft relative
from ...core.node.common.client import Client
from ...core.node.vm.plan_vm import PlanVirtualMachine
from ...core.pointer.pointer import Pointer
from .plan import Plan

PLAN_BUILDER_VM: PlanVirtualMachine = PlanVirtualMachine(name="plan_vm")
ROOT_CLIENT: Client = PLAN_BUILDER_VM.get_root_client()


def build_plan_inputs(forward_func: Callable) -> Dict[str, Pointer]:
    signature = inspect.signature(forward_func)
    res = {}
    for k, v in signature.parameters.items():
        if v.default is not inspect.Parameter.empty:
            res[k] = v.default.send(ROOT_CLIENT)
        else:
            raise ValueError(
                f"arg {k} has no placeholder as default value (required for @make_plan functions)"
            )
    return res


def map_in2out(inputs: dict, outputs: tuple) -> dict:
    in2out_map = {}
    for k, v in inputs.items():
        input_id_at_location = getattr(v, "id_at_location", None)
        for i, o in enumerate(outputs):
            if o.id_at_location == input_id_at_location:
                in2out_map[k] = i
    return in2out_map


def make_plan(func: Callable, inputs: Optional[Dict[str, Any]] = None) -> Plan:
    if inputs is None:
        inputs = build_plan_inputs(func)
    vm = PLAN_BUILDER_VM
    code = inspect.getsource(func)
    vm.record_actions()
    res = func(**inputs)
    vm.stop_recording()
    plan = Plan(
        actions=vm.recorded_actions,
        inputs=inputs,
        outputs=res,
        i2o_map=map_in2out(inputs, res),
        code=code,
    )
    # cleanup
    vm.recorded_actions = []
    return plan



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/lib/torch/module.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.plan import plan_pb2 as proto_dot_core_dot_plan_dot_plan__pb2
from syft.proto.lib.python import dict_pb2 as proto_dot_lib_dot_python_dot_dict__pb2
from syft.proto.lib.python import list_pb2 as proto_dot_lib_dot_python_dot_list__pb2
from syft.proto.lib.python.collections import (
    ordered_dict_pb2 as proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2,
)

DESCRIPTOR = _descriptor.FileDescriptor(
    name="proto/lib/torch/module.proto",
    package="syft.lib.torch",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x1cproto/lib/torch/module.proto\x12\x0esyft.lib.torch\x1a/proto/lib/python/collections/ordered_dict.proto\x1a\x1aproto/core/plan/plan.proto\x1a\x1bproto/lib/python/dict.proto\x1a\x1bproto/lib/python/list.proto"\xa7\x04\n\x06Module\x12\x13\n\x0bmodule_type\x18\x01 \x01(\t\x12\x13\n\x0bmodule_name\x18\x02 \x01(\t\x12\x18\n\x0bmodule_repr\x18\x03 \x01(\tH\x00\x88\x01\x01\x12\x15\n\rpython_module\x18\x04 \x01(\t\x12(\n\x08\x63hildren\x18\x05 \x03(\x0b\x32\x16.syft.lib.torch.Module\x12/\n\x0bmodule_args\x18\x06 \x01(\x0b\x32\x15.syft.lib.python.ListH\x01\x88\x01\x01\x12\x31\n\rmodule_kwargs\x18\x07 \x01(\x0b\x32\x15.syft.lib.python.DictH\x02\x88\x01\x01\x12<\n\nstate_dict\x18\x08 \x01(\x0b\x32(.syft.lib.python.collections.OrderedDict\x12<\n\nparameters\x18\t \x01(\x0b\x32(.syft.lib.python.collections.OrderedDict\x12*\n\x07\x66orward\x18\n \x01(\x0b\x32\x14.syft.core.plan.PlanH\x03\x88\x01\x01\x12@\n\t_uid2attr\x18\x0b \x01(\x0b\x32(.syft.lib.python.collections.OrderedDictH\x04\x88\x01\x01\x42\x0e\n\x0c_module_reprB\x0e\n\x0c_module_argsB\x10\n\x0e_module_kwargsB\n\n\x08_forwardB\x0c\n\nX_uid2attrb\x06proto3',
    dependencies=[
        proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2.DESCRIPTOR,
        proto_dot_core_dot_plan_dot_plan__pb2.DESCRIPTOR,
        proto_dot_lib_dot_python_dot_dict__pb2.DESCRIPTOR,
        proto_dot_lib_dot_python_dot_list__pb2.DESCRIPTOR,
    ],
)


_MODULE = _descriptor.Descriptor(
    name="Module",
    full_name="syft.lib.torch.Module",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="module_type",
            full_name="syft.lib.torch.Module.module_type",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="module_name",
            full_name="syft.lib.torch.Module.module_name",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="module_repr",
            full_name="syft.lib.torch.Module.module_repr",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="python_module",
            full_name="syft.lib.torch.Module.python_module",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="children",
            full_name="syft.lib.torch.Module.children",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="module_args",
            full_name="syft.lib.torch.Module.module_args",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="module_kwargs",
            full_name="syft.lib.torch.Module.module_kwargs",
            index=6,
            number=7,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="state_dict",
            full_name="syft.lib.torch.Module.state_dict",
            index=7,
            number=8,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="parameters",
            full_name="syft.lib.torch.Module.parameters",
            index=8,
            number=9,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="forward",
            full_name="syft.lib.torch.Module.forward",
            index=9,
            number=10,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="_uid2attr",
            full_name="syft.lib.torch.Module._uid2attr",
            index=10,
            number=11,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_module_repr",
            full_name="syft.lib.torch.Module._module_repr",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_module_args",
            full_name="syft.lib.torch.Module._module_args",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_module_kwargs",
            full_name="syft.lib.torch.Module._module_kwargs",
            index=2,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_forward",
            full_name="syft.lib.torch.Module._forward",
            index=3,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="X_uid2attr",
            full_name="syft.lib.torch.Module.X_uid2attr",
            index=4,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=184,
    serialized_end=735,
)

_MODULE.fields_by_name["children"].message_type = _MODULE
_MODULE.fields_by_name[
    "module_args"
].message_type = proto_dot_lib_dot_python_dot_list__pb2._LIST
_MODULE.fields_by_name[
    "module_kwargs"
].message_type = proto_dot_lib_dot_python_dot_dict__pb2._DICT
_MODULE.fields_by_name[
    "state_dict"
].message_type = (
    proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2._ORDEREDDICT
)
_MODULE.fields_by_name[
    "parameters"
].message_type = (
    proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2._ORDEREDDICT
)
_MODULE.fields_by_name[
    "forward"
].message_type = proto_dot_core_dot_plan_dot_plan__pb2._PLAN
_MODULE.fields_by_name[
    "_uid2attr"
].message_type = (
    proto_dot_lib_dot_python_dot_collections_dot_ordered__dict__pb2._ORDEREDDICT
)
_MODULE.oneofs_by_name["_module_repr"].fields.append(
    _MODULE.fields_by_name["module_repr"]
)
_MODULE.fields_by_name["module_repr"].containing_oneof = _MODULE.oneofs_by_name[
    "_module_repr"
]
_MODULE.oneofs_by_name["_module_args"].fields.append(
    _MODULE.fields_by_name["module_args"]
)
_MODULE.fields_by_name["module_args"].containing_oneof = _MODULE.oneofs_by_name[
    "_module_args"
]
_MODULE.oneofs_by_name["_module_kwargs"].fields.append(
    _MODULE.fields_by_name["module_kwargs"]
)
_MODULE.fields_by_name["module_kwargs"].containing_oneof = _MODULE.oneofs_by_name[
    "_module_kwargs"
]
_MODULE.oneofs_by_name["_forward"].fields.append(_MODULE.fields_by_name["forward"])
_MODULE.fields_by_name["forward"].containing_oneof = _MODULE.oneofs_by_name["_forward"]
_MODULE.oneofs_by_name["X_uid2attr"].fields.append(_MODULE.fields_by_name["_uid2attr"])
_MODULE.fields_by_name["_uid2attr"].containing_oneof = _MODULE.oneofs_by_name[
    "X_uid2attr"
]
DESCRIPTOR.message_types_by_name["Module"] = _MODULE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Module = _reflection.GeneratedProtocolMessageType(
    "Module",
    (_message.Message,),
    {
        "DESCRIPTOR": _MODULE,
        "__module__": "proto.lib.torch.module_pb2"
        # @@protoc_insertion_point(class_scope:syft.lib.torch.Module)
    },
)
_sym_db.RegisterMessage(Module)


# @@protoc_insertion_point(module_scope)



# stdlib
import ast
from collections import OrderedDict
import copy
import importlib
import inspect
from itertools import islice
import os
from pathlib import Path
import sys
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import torch

# syft absolute
import syft as sy
from syft.core.node.common.action.save_object_action import SaveObjectAction
from syft.core.plan.plan_builder import ROOT_CLIENT
from syft.core.plan.plan_builder import make_plan
from syft.lib.python import _SyNone

# syft relative
from .. import lib_ast
from ...core.pointer.pointer import Pointer
from ...generate_wrapper import GenerateWrapper
from ...lib.util import full_name_with_qualname
from ...logger import critical
from ...logger import info
from ...logger import traceback_and_raise
from ...proto.lib.torch.module_pb2 import Module as Module_PB
from ..python.collections import OrderedDict as SyOrderedDict
from ..python.util import downcast
from ..python.util import upcast

# from ...core.node.common.service.auth import AuthorizationException


def repr_to_kwargs(repr_str: str) -> Tuple[List[Any], Dict[Any, Any]]:
    # for example: repr_str = Conv2d(...).extra_repr()
    # produces:    > str("1, 32, kernel_size=(3, 3), stride=(1, 1)")
    # then we just have to split it into args and kwargs
    # using ast.literal_eval we can use python to give us the real primitive types
    # from the strings in a safe way
    # str("1 ") becomes int(1)
    # str("(1, 2) ") becomes tuple(1, 2)
    args: List[Any] = []
    kwargs: Dict[Any, Any] = {}
    parts = repr_str.split(",")

    # tuples are split by commas as well, so we will keep a tab on open parentheses
    # then concat with "," until we find a close parentheses
    # TODO: make work nested with a count and add tests
    para_open = False
    buffer = ""
    for part in parts:
        try:
            if "(" in part:
                para_open = True
                buffer = ""
            if para_open is True:
                buffer += part + ","
                if ")" in part:
                    # remove trailing ,
                    part = buffer[:-1]
                    buffer = ""
                    para_open = False
                else:
                    continue

            string = part.strip()
            if "=" not in string:
                # its an arg
                arg = ast.literal_eval(string)
                args.append(arg)
            else:
                # its a kwarg
                kv = string.split("=")
                key = str(kv[0])
                string = kv[1].strip()
                value = ast.literal_eval(string)
                kwargs[key.strip()] = value
        except Exception as e:
            info(f"ast.literal_eval failed to parse part: {string}. {e}")

    return (args, kwargs)


class Module:
    """
    This is our equivalent of torch.nn.Module and aims to have the same external
    interface. We will need to support both torch Modules and Module Pointers.
    """

    def __init__(self, torch_ref: Any) -> None:
        self.setup(torch_ref=torch_ref)

    def setup(self, torch_ref: Any) -> None:
        # the remote torch means the model is remote
        self.remote_model: Optional["Module"] = None
        self.local_model: Optional["Module"] = None
        self.duet = None
        if "syft" in full_name_with_qualname(klass=type(torch_ref)):
            info("> Creating remote model")
            self.is_local = False
        else:
            # otherwise we have a local model
            info("> Creating local model")
            self.is_local = True

        self.torch_ref = torch_ref
        self.training = False
        self._modules: OrderedDict[str, Module] = OrderedDict()
        real_module = torch_ref.nn.Module()
        self.__dict__["real_module"] = real_module  # bypass getattr/setattr
        # if issubclass(type(real_module), Pointer):
        #     try:
        #         # TODO: this needs fixing but should be on by default for now
        #         # https://github.com/OpenMined/PySyft/issues/5242
        #         real_module.searchable = True
        #     except AuthorizationException as e:
        #         print(f"Cant make real_module searchable. {e}")

    def __setattr__(self, name: str, value: Union[Any, "Module"]) -> None:
        # this is how we catch the modules being set during subclass init
        # bug where torch.nn.modules isn't the full name on some imports
        # TODO: fix this properly
        # third party
        import torch

        if "torch.nn" in full_name_with_qualname(klass=type(value)) or isinstance(
            value, torch.nn.Module
        ):
            modules = self.__dict__.get("_modules")
            if modules is not None:
                modules[name] = value

            # attach all the sub modules to a real module so that we can have a
            # remote module pointer that acts like a real model
            real_module: Optional[OrderedDict] = self.__dict__.get("real_module")
            if real_module is not None:
                real_module.add_module(name, value)  # type: ignore
        else:
            object.__setattr__(self, name, value)

    def __getattr__(self, name: str) -> Union[Any, "Module"]:
        modules: Optional[OrderedDict] = self.__dict__.get("_modules")
        if modules is not None:
            if name in modules:
                return modules[name]

        return object.__getattribute__(self, name)

    def train(self, mode: bool = True) -> "Module":
        self.training = mode
        for _, module in self.modules.items():
            module.train(mode)
        return self

    def eval(self) -> "Module":
        return self.train(False)

    def __call__(
        self, *args: Union[List[Any], Tuple[Any, ...]], **kwargs: Dict[Any, Any]
    ) -> Any:
        return self.forward(*args, **kwargs)

    @property
    def modules(self) -> OrderedDict:
        modules = self.__dict__.get("_modules")
        if modules is not None:
            return modules
        return OrderedDict()

    # local list of remote ListPointers of TensorPointers
    def parameters(self, recurse: bool = True) -> Optional[List[Any]]:
        params_list: Optional[List[Any]] = None

        if self.is_local is True:
            # we are local so use normal torch params
            params_list = []
        for _, module in self.modules.items():
            params = module.parameters(recurse)
            if params_list is None:
                # only on remote create a remote list so we can concat the param list
                # pointers without having to actually get them
                self.duet = params.client
                params_list = self.duet.syft.lib.python.List()  # type: ignore
            # either way lets concat them until we have a big list of parameters
            params_list += params
        return params_list

    def cuda(self, device: Any) -> "Module":
        for _, module in self.modules.items():
            module.cuda(device)
        return self

    def cpu(self) -> "Module":
        for _, module in self.modules.items():
            module.cpu()
        return self

    def load_state_dict(self, input: Union[str, os.PathLike, Dict[str, Any]]) -> None:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return None

        state_dict = {}
        if isinstance(input, (str, os.PathLike)):
            with open(Path(input), "rb") as f:
                state_dict = torch.load(f)
        else:
            state_dict = dict(input)

        if not issubclass(type(state_dict), dict):
            traceback_and_raise(
                f"  Invalid input: {type(input)}. "
                + "Try inputting a state_dict or .pth file."
            )

        info("> Loading model weights")
        layers: Dict[str, Any] = {}
        for save_key, values in state_dict.items():
            parts = save_key.split(".")
            if len(parts) < 2:
                info(f"  state dict key is too short: {save_key}")
                continue
            layer = parts[0]
            attr = parts[1]
            if layer not in layers:
                layers[layer] = {}
            layers[layer][attr] = values

        for layer, sd in layers.items():
            local_layer = getattr(self, layer, None)
            if local_layer is not None and hasattr(local_layer, "load_state_dict"):
                d = local_layer.load_state_dict(sd)
                info(f"  {layer} state dict loaded with: {d}")
            else:
                info(f"  Model doesnt have layer {layer}")

        info("> Finished loading weights")
        return None

    def state_dict(self) -> Optional[Dict[str, Any]]:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return None

        info("> Saving model weights")
        model_state_dict = OrderedDict()
        for name, module in self.modules.items():
            if hasattr(module, "state_dict"):
                for k, v in module.state_dict().items():
                    save_key = f"{name}.{k}"
                    model_state_dict[save_key] = v

        info("> Finished saving weights")
        return model_state_dict

    def save(self, path: Union[str, bytes, os.PathLike]) -> None:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return

        state_dict = self.state_dict()
        torch.save(state_dict, path)

    def load(self, path: Union[str, os.PathLike]) -> None:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return

        self.load_state_dict(input=path)

    def send(self, client: Any, send_parameters: bool = True) -> Any:
        if not self.is_local:
            info("> This model is remote so try calling .get()")
            return

        info("> Sending local model")

        remote_model = copy.copy(self)
        remote_model.setup(torch_ref=client.torch)
        remote_model.duet = client

        for name, module in self.modules.items():
            fqn = full_name_with_qualname(klass=type(module))
            klass = client.lib_ast.query(fqn, obj_type=type(module))
            module_repr = module.extra_repr()
            args, kwargs = repr_to_kwargs(repr_str=module_repr)
            remote_module_ptr = klass(*args, **kwargs)
            remote_model.__setattr__(name, remote_module_ptr)

            # if the remote module has state_dict lets get it
            if (
                send_parameters
                and hasattr(module, "state_dict")
                and hasattr(remote_module_ptr, "load_state_dict")
            ):
                local_state_ord_dict = module.state_dict()
                # cast to dict because OrderedDict is not supported

                # get a blocking copy of the state_dict
                info(f"  Sending local layer: {name}")
                # cant import Dict / PrimitiveFactory due to circular imports
                remote_state_dict_ptr = client.syft.lib.python.Dict(
                    dict(local_state_ord_dict)
                )
                # iterate through the key, values
                # weights and biases should be in there
                remote_module_ptr.load_state_dict(remote_state_dict_ptr)

        info("\n> Finished sending local model <\n\n")
        self.remote_model = remote_model
        return self.remote_model

    def get(
        self,
        request_block: bool = False,
        timeout_secs: int = 20,
        reason: str = "",
        delete_obj: bool = False,
    ) -> Optional["Module"]:

        if self.is_local:
            info("> This model is local. Maybe you meant to call .send()?")
            return None

        info("> Downloading remote model")

        local_model = copy.copy(self)
        local_model.setup(torch_ref=torch)
        local_model.duet = self.duet

        for layer_name, module in self.modules.items():
            module_parts = module.path_and_name.split(".")
            klass_name = module_parts.pop()
            klass = getattr(sys.modules[".".join(module_parts)], klass_name)
            repr_ptr = module.extra_repr()

            module_repr = repr_ptr.get(
                request_block=request_block,
                reason=reason,
                timeout_secs=timeout_secs,
            )

            if module_repr is None:
                info(f"  Request for {reason} extra_repr failed, skipping layer")
                continue

            args, kwargs = repr_to_kwargs(repr_str=module_repr.upcast())
            local_module = klass(*args, **kwargs)

            # the local real module has been set on the sy module
            local_model.__setattr__(layer_name, local_module)

            try:
                # if the remote module has state_dict lets get it
                if hasattr(module, "state_dict") and hasattr(
                    local_module, "load_state_dict"
                ):
                    info("loading remote state dict")
                    sd_ptr = module.state_dict()
                    # get a blocking copy of the state_dict
                    info(f"  Downloading remote layer: {layer_name}")
                    state_dict = sd_ptr.get(
                        request_block=request_block,
                        reason=reason,
                        timeout_secs=timeout_secs,
                        delete_obj=delete_obj,
                    )
                    # We have to recreate the OrderedDict for load_state_dict to work
                    ordered_state_dict = OrderedDict()
                    for elem, item in state_dict.items():
                        ordered_state_dict[str(elem)] = item
                    # iterate through the key, values
                    # weights and biases should be in there
                    if state_dict is not None:
                        # TODO: support torch.nn.modules.module._IncompatibleKeys
                        local_module.load_state_dict(ordered_state_dict)
                    else:
                        info(
                            f"  Failed to get {layer_name} state_dict, skipping layer."
                        )

            except Exception as e:
                critical(f"  Failed to download remote state for {layer_name}.")
                traceback_and_raise(e)

        info("\n> Finished downloading remote model <\n\n")
        self.local_model = local_model
        return self.local_model

    # zero them so we know they are copied
    def zero_layers(self) -> None:
        for m in self.modules.values():
            if hasattr(m, "weight"):
                m.weight.requires_grad_(False).zero_()
            if hasattr(m, "bias"):
                m.bias.requires_grad_(False).zero_()

    # easy way to check the weights have changed
    def debug_sum_layers(self) -> None:
        info("> Summing layers for debugging: ")
        for n, m in self.modules.items():
            if hasattr(m, "state_dict"):
                if self.is_local:
                    state_dict = m.state_dict()
                else:
                    state_dict = m.state_dict().get()

                for k, v in state_dict.items():
                    if hasattr(v, "sum"):
                        s = v.sum().item()
                        info(f"  Layer {n} sum({k}): {s}")


def module_in_ast(module_type: type) -> bool:
    fqn = full_name_with_qualname(klass=module_type)
    try:
        return bool(lib_ast.query(fqn, obj_type=module_type))
    except ValueError:
        return False


def object2proto(obj: torch.nn.Module, is_child: bool = False) -> Module_PB:

    proto = Module_PB()
    if module_in_ast(type(obj)):
        proto.python_module = type(obj).__module__
    else:
        proto.python_module = "_USER_DEFINED"
        proto.forward.CopyFrom(sy.serialize(obj._forward_plan))

    proto.module_type = type(obj).__name__
    proto.module_repr = obj.extra_repr()

    # Requires klass.store_init_args() when building AST.
    if hasattr(obj, "_init_args") and hasattr(obj, "_init_kwargs"):
        module_args = sy.serialize(sy.lib.python.List(obj._init_args))
        proto.module_args.CopyFrom(module_args)
        module_kwargs = sy.serialize(sy.lib.python.Dict(obj._init_kwargs))
        proto.module_kwargs.CopyFrom(module_kwargs)

    if hasattr(obj, "_uid2attr"):
        proto._uid2attr.CopyFrom(sy.serialize(SyOrderedDict(obj._uid2attr)))

    proto.parameters.CopyFrom(sy.serialize(SyOrderedDict(obj._parameters)))

    for n, m in obj.named_children():
        child_proto = object2proto(m, is_child=True)
        child_proto.module_name = n
        proto.children.append(child_proto)

    return proto


def proto2object(proto: Module_PB) -> torch.nn.Module:
    is_userdefined = proto.python_module == "_USER_DEFINED"

    if is_userdefined:
        obj_type = type(
            proto.module_type,
            (torch.nn.Module,),
            {},
        )
    else:
        obj_type = getattr(
            importlib.import_module(proto.python_module), proto.module_type
        )

    if proto.HasField("module_args") and proto.HasField("module_kwargs"):
        args = upcast(sy.deserialize(proto.module_args))
        kwargs = upcast(sy.deserialize(proto.module_kwargs))
    elif proto.HasField("module_repr"):
        # TODO make Modules defined by torch use module_args and module_kwargs,
        # and remove module_repr from proto
        args, kwargs = repr_to_kwargs(repr_str=proto.module_repr)
    else:
        raise ValueError("Could not infer args and kwargs from Module proto.")

    obj = obj_type(*args, **kwargs)

    for name, param in sy.deserialize(proto.parameters).items():
        # if we don't do this check, some torch.nn layers fail ( e.g. Conv2D with bias=False)
        if not isinstance(param, _SyNone):
            setattr(obj, str(name), param)

    if proto.HasField("forward"):
        forward_plan = sy.deserialize(proto.forward)
        obj._forward_plan = forward_plan
        compile_and_forward = create_compile_and_forward_fn(obj)
        obj.__call__ = compile_and_forward
        obj.forward = compile_and_forward
        # obj.__call__ = forward_plan
        # obj.forward = forward_plan

    for child_proto in proto.children:
        setattr(obj, str(child_proto.module_name), sy.deserialize(child_proto))

    if proto.HasField("_uid2attr"):
        obj._uid2attr = sy.deserialize(proto._uid2attr)

    if is_userdefined:
        recompile(obj)

    return obj


def create_compile_and_forward_fn(obj: "SyModule") -> Callable:
    """Wraps a forward plan in a function that first recompiles the plan, and then
    executes the plan

    Args:
        obj (SyModule): the SyModule
    """

    def _compile_and_forward(*args, **kwargs):  # type: ignore
        recompile(obj)
        return obj._forward_plan(*args, **kwargs)

    return _compile_and_forward


def recompile(sy_module: "SyModule") -> None:
    """Recompiles the forward plan, if the object state has changed since the
    forward plan was created, we update the plan here

    Args:
        sy_module (SyModule): the module to compile
    """
    if hasattr(sy_module, "_forward_plan"):
        for action in sy_module._forward_plan.actions:  # type: ignore
            if (
                isinstance(action, SaveObjectAction)
                and action.obj.id in sy_module._uid2attr
            ):
                action.obj.data = getattr(
                    sy_module, str(sy_module._uid2attr[action.obj.id])
                )


GenerateWrapper(
    wrapped_type=torch.nn.Module,
    import_path="torch.nn.Module",
    protobuf_scheme=Module_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)


class ForwardToPlanConverter(type):
    """This metaclass ensures that:
    1) the object is initialized when calling Object()
    2) obj._make_forward_plan() is called after initialization
    """

    def __call__(cls: Any, *args, **kwargs) -> Any:  # type: ignore
        # TODO: check if contains input_size
        obj = type.__call__(cls, *args, **kwargs)
        obj._make_forward_plan()
        return obj


class SyModule(torch.nn.Module, metaclass=ForwardToPlanConverter):
    """A `SyModule` is the pointable equivalent of a torch.nn.Module. In order to make
    SyModule remotely executable, its `.forward` method is converted into a `Plan` object
    when initializing a `SyModule` object. This object has two "modes", in which it behaves
    differently. During the "forward plan building stage" it transforms parameters and submodules
    into pointer when the user retrieves them. After plan building the model behaves more
    like a regular torch.nn.Module, but instead of running a forward method, the user executes
    a `Plan`. As the user does not need to understand the building stage, and the .forward API
    is fairly similar to a regular torch.nn.Module, there is no need to understand all internals
    to use this module.

    """

    def __init__(  # type: ignore
        self,
        *args,
        input_size: Optional[Tuple[int]] = None,
        inputs: Optional[Dict[str, torch.Tensor]] = None,
        **kwargs,
    ):
        """
        Args:
            input_size (Optional[Tuple[int]], optional): input_size of the Module,
                needs to be defined or inferrable. Defaults to None.
            inputs (Optional[Dict[str, torch.Tensor]], optional): dictionary of dummy input tensors.
                Use this argument instead of `input_size` if there are multiple forward inputs,
                or if the forward input is not a FloatTensor. Defaults to None.

        Raises:
            ValueError: [description]
        """
        super().__init__(*args, **kwargs)
        self.building_forward = False
        self._parameter_pointers: Dict[str, Pointer] = dict()

        if (input_size is None) == (inputs is None):
            raise ValueError(
                "Either `input_size` or `inputs` should be specified, but not both."
            )

        self.input_size = input_size
        self.inputs = inputs

    def _make_forward_plan(self) -> None:
        """Convert forward function into a `Plan` object

        Raises:
            ValueError: `.forward` method must be defined
        """
        if getattr(self.forward, __name__, None) == "_forward_unimplemented":  # type: ignore
            raise ValueError("Missing .forward() method for Module")

        inputs = self._get_forward_inputs()

        self.building_forward = True
        plan = make_plan(self.forward, inputs=inputs)  # type: ignore
        self.forward = self._local_forward
        self._forward_plan = plan
        self.__call__ = plan
        self._create_uid2attr()
        self.building_forward = False
        self._remove_plan_action_data()

    def _remove_plan_action_data(self) -> None:
        """
        Sets `action.obj.data` for each symodule action in `self._forward_plan` to `None`.

        This greatly reduces the proto memory footprint;
        The whole state of `self` is saved in the action, which will be recompiled anyway.
        """

        # Remove module action data
        for action in self._forward_plan.actions:
            if isinstance(action, SaveObjectAction) and action.obj.id in self._uid2attr:
                action.obj.data = downcast(None)

    def _local_forward(self, *args, **kwargs):  # type: ignore
        recompile(self)
        return self._forward_plan(*args, **kwargs)

    def _create_uid2attr(self) -> None:
        self._uid2attr = {
            param.id_at_location: attr_name
            for attr_name, param in self._parameter_pointers.items()
        }

    def __getattr__(self, name: str) -> Any:
        """A custom getattr method. When retrieving a torch.nn.Module or a torch.nn.Parameter
        *during forward plan building*, SyModule instead returns a Pointer to this attribute.
        The first time an attribute is retrieved, we send it to the plan builder VM, and store
        it in self._parameters_pointers, which will be used for plan Recompilation during
        *deserialization*. If an attribute is requested again, we return the pointer from
        `_parameters_pointers`

        Args:
            name (str): name of the attr

        Returns:
            Any: Attribute value or Pointer to it
        """
        # this is __getattr__ instead of __getattribute__ because of the structure of torch.nn.Module
        if name in self._parameter_pointers and self.building_forward:
            return self._parameter_pointers[name]

        res = super().__getattr__(name)
        if (
            isinstance(res, (torch.nn.Module, torch.nn.Parameter))
            and self.building_forward
        ):
            res_ptr = res.send(ROOT_CLIENT)
            self._parameter_pointers[name] = res_ptr
            return res_ptr
        else:
            return res

    def _get_inp_keys(self) -> List[str]:
        """Get key for the `.forward` argument
        Returns:
            str: input key
        """

        forward_signature = inspect.signature(self.forward)
        args = list(forward_signature.parameters.items())
        if len(args) == 0:
            raise ValueError("SyModules requires more than one argument, and no kwargs")
        inp_keys = []
        for k, v in args:
            if v.default is not inspect.Parameter.empty:
                raise ValueError("SyModules accept only args, not kwargs")
            inp_keys.append(k)
        return inp_keys

    def _get_forward_inputs(self) -> Dict[str, Pointer]:
        """Get the dummy inputs for generating the .forward `Plan`

        Returns:
            Dict[str: Any]: inputs for .forward
        """
        inp_keys = self._get_inp_keys()

        if hasattr(self, "inputs") and isinstance(self.inputs, dict):
            if set(inp_keys) != set(self.inputs.keys()):
                raise ValueError(
                    "Given `inputs` dict and expected `forward` inputs do not match."
                )
            inputs = {k: v.send(ROOT_CLIENT) for k, v in self.inputs.items()}

        elif hasattr(self, "input_size") and isinstance(self.input_size, tuple):
            if len(inp_keys) != 1:
                raise ValueError(
                    "`.forward` method has more than one input, define dummy inputs with `inputs` kwarg."
                )
            inputs = {inp_keys[0]: torch.rand(self.input_size).send(ROOT_CLIENT)}

        else:
            raise ValueError(
                "SyModule needs either `input_size`: Tuple(Int) or `inputs`: Dict[str, Tensor] as kwarg"
                "to trace the forward plan."
                "Also, make sure to call **super().__init__(**kwargs)** in ALL your SyModules"
                ""
            )

        return inputs


class SySequential(SyModule):
    """The Syft equivalent of torch.nn.Sequential"""

    def __init__(self, *args, input_size: Optional[Tuple[int]] = None, **kwargs):  # type: ignore
        """initializes SySequential and stores the submodules

        input_size (Tuple[Int], optional): input_size of the Module, needs to be defined or inferrable.
            Defaults to None.
        """
        if input_size is None:
            input_size = self._infer_input_size(*args)

        super().__init__(input_size=input_size, **kwargs)
        for idx, module in enumerate(args):
            setattr(self, str(idx), module)
        self.n_modules = len(args)

    def _infer_input_size(self, *args: Any) -> Tuple[int]:
        """Infer input size from first child

        Returns:
            Tuple[int]: input size of first child SyModule.
        """
        if hasattr(args[0], "input_size"):
            return args[0].input_size
        else:
            raise ValueError(
                "Could not infer `input_size` from children modules."
                "Either 1) define `input_size` as a kwarg of SySequential OR"
                " 2) define `input_size`: Tuple[int] as kwarg on the first child module."
            )

    def __iter__(self):  # type: ignore
        if self.building_forward:
            return iter([getattr(self, str(i)) for i in range(self.n_modules)])
        else:
            return iter(self._modules.values())

    def _get_item_by_idx(self, iterator: Iterator, idx: int) -> SyModule:
        """Get the idx-th item of the iterator"""
        size = self.n_modules
        if not -size <= idx < size:
            raise IndexError(f"index {idx} is out of range")
        return next(islice(iterator, idx, None))

    def __getitem__(self, idx: int) -> SyModule:
        if isinstance(idx, slice):  # type: ignore
            raise ValueError("SySequential does not support slices")
        else:
            return self._get_item_by_idx(self._modules.values(), idx)

    def __setitem__(self, idx: int, module: Module) -> None:
        key = self._get_item_by_idx(self._modules.keys(), idx)
        return setattr(self, key, module)

    def __delitem__(self, idx: Union[slice, int]) -> None:
        if isinstance(idx, slice):
            raise ValueError("SySequential does not support slices")
        else:
            key = self._get_item_by_idx(self._modules.keys(), idx)
            delattr(self, key)

    def forward(self, x: Any) -> Any:  # type: ignore
        """Sequentially call submodule.forward

        Args:
            x (Any, optional): input. Defaults to None.

        Returns:
            Any: Module output
        """
        out = x
        for i, module in enumerate(self):
            # handle indexing in the block, or in the sequential?
            if module.__class__.__name__ == "ModulePointer":
                # user defined module
                out = module.forward(x=out)[0]
            else:
                # AST module
                out = module(out)
        return out

    def _get_inp_key(self) -> str:
        """Get key for the `.forward` argument, allways x for this module

        Returns:
            str: "x"
        """
        return "x"



# third party
import pandas as pd

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.python.list import List
from ...lib.python.primitive_factory import PrimitiveFactory
from ...proto.lib.pandas.categorical_pb2 import (
    PandasCategoricalDtype as PandasCategoricalDtype_PB,
)


def object2proto(obj: pd.CategoricalDtype) -> PandasCategoricalDtype_PB:
    # since pd.Index type is not integrated converted obj.categories to List
    pd_cat_list = PrimitiveFactory.generate_primitive(value=obj.categories.tolist())
    cat_list_proto = pd_cat_list._object2proto()

    return PandasCategoricalDtype_PB(
        id=cat_list_proto.id, categories=cat_list_proto, ordered=obj.ordered
    )


def proto2object(proto: PandasCategoricalDtype_PB) -> pd.CategoricalDtype:
    categories = List._proto2object(proto.categories).upcast()
    ordered = proto.ordered
    return pd.CategoricalDtype(categories=categories, ordered=ordered)


GenerateWrapper(
    wrapped_type=pd.CategoricalDtype,
    import_path="pandas.CategoricalDtype",
    protobuf_scheme=PandasCategoricalDtype_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



"""Serde method for pd.DataFrame."""

# third party
import pandas as pd
import pyarrow as pa

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...proto.lib.pandas.frame_pb2 import PandasDataFrame as PandasDataFrame_PB


def object2proto(obj: pd.DataFrame) -> PandasDataFrame_PB:
    """Convert pd.DataFrame to PandasDataFrame_PB with pyarrow.

    Args:
        obj: target Dataframe

    Returns:
        Serialized version of Dataframe, which will be used to reconstruction.

    """
    schema = pa.Schema.from_pandas(obj)
    table = pa.Table.from_pandas(obj)
    sink = pa.BufferOutputStream()

    writer = pa.ipc.new_file(sink, schema)
    writer.write(table)
    writer.close()

    buf = sink.getvalue()

    siz = len(buf)
    df_bytes = pa.compress(buf, asbytes=True)

    return PandasDataFrame_PB(dataframe=df_bytes, decompressed_size=siz)


def proto2object(proto: PandasDataFrame_PB) -> pd.DataFrame:
    """Proto to object conversion using to return desired model.

    Args:
        proto: Serialized version of Dataframe, which will be used to reconstruction.

    Returns:
        Re-constructed dataframe.
    """
    buf = pa.decompress(proto.dataframe, decompressed_size=proto.decompressed_size)
    return pa.ipc.open_file(buf).read_pandas()


GenerateWrapper(
    wrapped_type=pd.DataFrame,
    import_path="pandas.DataFrame",
    protobuf_scheme=PandasDataFrame_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import statsmodels
import statsmodels.api as sm

# syft relative
from . import family  # noqa: 401
from . import results  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "statsmodels"
PACKAGE_SUPPORT = {"lib": LIB_NAME}


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client=client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("statsmodels", statsmodels),
        ("statsmodels.api", sm),
        ("statsmodels.genmod", statsmodels.genmod),
        (
            "statsmodels.genmod.generalized_linear_model",
            statsmodels.genmod.generalized_linear_model,
        ),
        ("statsmodels.genmod.families", statsmodels.genmod.families),
        ("statsmodels.iolib", statsmodels.iolib),
        ("statsmodels.iolib.summary", statsmodels.iolib.summary),
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        (
            "statsmodels.genmod.generalized_linear_model.GLM",
            "statsmodels.genmod.generalized_linear_model.GLM",
            statsmodels.genmod.generalized_linear_model.GLM,
        ),
        (
            "statsmodels.genmod.generalized_linear_model.GLMResults",
            "statsmodels.genmod.generalized_linear_model.GLMResults",
            statsmodels.genmod.generalized_linear_model.GLMResults,
        ),
        (
            "statsmodels.genmod.generalized_linear_model.GLMResultsWrapper",
            "statsmodels.genmod.generalized_linear_model.GLMResultsWrapper",
            statsmodels.genmod.generalized_linear_model.GLMResultsWrapper,
        ),
        (
            "statsmodels.iolib.summary.Summary",
            "statsmodels.iolib.summary.Summary",
            statsmodels.iolib.summary.Summary,
        ),
        (
            "statsmodels.genmod.families.Binomial",
            "statsmodels.genmod.families.Binomial",
            statsmodels.genmod.families.Binomial,
        ),
        (
            "statsmodels.genmod.families.Gamma",
            "statsmodels.genmod.families.Gamma",
            statsmodels.genmod.families.Gamma,
        ),
        (
            "statsmodels.genmod.families.Gaussian",
            "statsmodels.genmod.families.Gaussian",
            statsmodels.genmod.families.Gaussian,
        ),
        (
            "statsmodels.genmod.families.InverseGaussian",
            "statsmodels.genmod.families.InverseGaussian",
            statsmodels.genmod.families.InverseGaussian,
        ),
        (
            "statsmodels.genmod.families.NegativeBinomial",
            "statsmodels.genmod.families.NegativeBinomial",
            statsmodels.genmod.families.NegativeBinomial,
        ),
        (
            "statsmodels.genmod.families.Poisson",
            "statsmodels.genmod.families.Poisson",
            statsmodels.genmod.families.Poisson,
        ),
        (
            "statsmodels.genmod.families.Tweedie",
            "statsmodels.genmod.families.Tweedie",
            statsmodels.genmod.families.Tweedie,
        ),
    ]

    # TODO: finish all these methods, summary is an object attribute for example
    methods = [
        ("statsmodels.api.add_constant", "pandas.DataFrame"),
        (
            "statsmodels.genmod.generalized_linear_model.GLM.fit",
            "statsmodels.genmod.generalized_linear_model.GLMResultsWrapper",
        ),
        (
            "statsmodels.genmod.generalized_linear_model.GLMResults.summary",
            "statsmodels.iolib.summary.Summary",
        ),
        (
            "statsmodels.genmod.generalized_linear_model.GLMResultsWrapper.cov_params",
            "pandas.DataFrame",
        ),
        (
            "statsmodels.genmod.generalized_linear_model.GLMResultsWrapper.conf_int",
            "pandas.DataFrame",
        ),
        ("statsmodels.iolib.summary.Summary.as_csv", "syft.lib.python.String"),
        ("statsmodels.iolib.summary.Summary.as_html", "syft.lib.python.String"),
        ("statsmodels.iolib.summary.Summary.as_latex", "syft.lib.python.String"),
        ("statsmodels.iolib.summary.Summary.as_text", "syft.lib.python.String"),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



"""Partial-dependency library, runs when user loads sklearn.

__init__ file for sklearn. This defines various modules, classes and methods which we currently support.
We create an AST for all these modules, classes and methods so that they can be called remotely.
"""

# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import sklearn
import sklearn.linear_model

# syft relative
from . import serializing_models  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "sklearn"
PACKAGE_SUPPORT = {"lib": LIB_NAME}


def create_ast(client: TypeAny = None) -> Globals:
    """Create ast for all mdules, classes and attributes for sklearn so that they can be called remotely.

    Args:
        client: Remote client where we have to create the ast.

    Returns:
        ast created for sklearn.
    """
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("sklearn", sklearn),
        ("sklearn.linear_model", sklearn.linear_model),
    ]

    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        ("sklearn.base", "sklearn.base", sklearn.base),
        # linear_model
        # LogisticRegression
        (
            "sklearn.linear_model.LogisticRegression",
            "sklearn.linear_model.LogisticRegression",
            sklearn.linear_model.LogisticRegression,
        ),
        # LinearRegression
        (
            "sklearn.linear_model.LinearRegression",
            "sklearn.linear_model._base.LinearRegression",
            sklearn.linear_model._base.LinearRegression,
        ),
    ]

    methods: TypeList[TypeTuple[str, str]] = [
        # linear_model
        # LogisticRegression
        (
            "sklearn.linear_model.LogisticRegression.fit",
            "sklearn.linear_model.LogisticRegression",
        ),
        ("sklearn.linear_model.LogisticRegression.predict", "pandas.DataFrame"),
        # LinearRegression
        (
            "sklearn.linear_model.LinearRegression.fit",
            "sklearn.linear_model._base.LinearRegression",
        ),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



# stdlib
from typing import Dict
from typing import Type

# third party
import statsmodels

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.python.primitive_factory import PrimitiveFactory
from ...lib.python.string import String
from ...proto.lib.statsmodels.family_pb2 import FamilyProto

FAMILY_2_STR: Dict[Type[statsmodels.genmod.families.family.Family], str] = {
    statsmodels.genmod.families.family.Binomial: "Binomial",
    statsmodels.genmod.families.family.Gamma: "Gamma",
    statsmodels.genmod.families.family.Gaussian: "Gaussian",
    statsmodels.genmod.families.family.InverseGaussian: "InverseGaussian",
    statsmodels.genmod.families.family.NegativeBinomial: "NegativeBinomial",
    statsmodels.genmod.families.family.Poisson: "Poisson",
    statsmodels.genmod.families.family.Tweedie: "Tweedie",
}
LINK_2_STR: Dict[Type[statsmodels.genmod.families.family.Family], str] = {
    statsmodels.genmod.families.links.log: "log",
    statsmodels.genmod.families.links.logit: "logit",
    statsmodels.genmod.families.links.cauchy: "cauchy",
    statsmodels.genmod.families.links.cloglog: "cloglog",
    statsmodels.genmod.families.links.identity: "identity",
    statsmodels.genmod.families.links.inverse_power: "inverse_power",
    statsmodels.genmod.families.links.inverse_squared: "inverse_squared",
    statsmodels.genmod.families.links.nbinom: "nbinom",
    statsmodels.genmod.families.links.probit: "probit",
    statsmodels.genmod.families.links.Power: "Power",
    statsmodels.genmod.families.links.NegativeBinomial: "NegativeBinomial",
    statsmodels.genmod.families.links.CDFLink: "CDFLink",
    statsmodels.genmod.families.links.sqrt: "sqrt",
}

STR_2_FAMILY: Dict[str, Type[statsmodels.genmod.families.family.Family]] = {
    v: k for k, v in FAMILY_2_STR.items()
}
STR_2_LINK: Dict[str, Type[statsmodels.genmod.families.family.Family]] = {
    v: k for k, v in LINK_2_STR.items()
}


def object2proto(obj: Type[statsmodels.genmod.families.family.Family]) -> FamilyProto:
    family_name = FAMILY_2_STR[type(obj)]
    link_name = LINK_2_STR[type(obj.link)]
    family_name_prim = PrimitiveFactory.generate_primitive(value=family_name)
    link_name_prim = PrimitiveFactory.generate_primitive(value=link_name)
    family_name_proto = family_name_prim._object2proto()
    link_name_proto = link_name_prim._object2proto()

    return FamilyProto(family=family_name_proto, link=link_name_proto)


def proto2object(proto: FamilyProto) -> Type[statsmodels.genmod.families.family.Family]:
    family_name = str(String._proto2object(proto=proto.family))
    link_name = str(String._proto2object(proto=proto.link))
    obj = STR_2_FAMILY[family_name](link=STR_2_LINK[link_name])
    return obj


for fam in FAMILY_2_STR.keys():
    GenerateWrapper(
        wrapped_type=fam,
        import_path="statsmodels.genmod.families.family" + fam.__class__.__name__,
        protobuf_scheme=FamilyProto,
        type_object2proto=object2proto,
        type_proto2object=proto2object,
    )



"""add the sympc library into syft."""

# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# syft relative
from . import rst_share  # noqa: 401
from . import session  # noqa: 401
from . import share  # noqa: 401
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "sympc"
PACKAGE_SUPPORT = {
    "lib": LIB_NAME,
    "torch": {"min_version": "1.6.0", "max_version": "1.8.1"},
    "python": {"min_version": (3, 7), "max_version": (3, 9, 99)},
}


def create_ast(client: TypeAny = None) -> Globals:
    """Add the modules, classes and attributes from sympc to syft.

    Args:
        client: Client

    Returns:
        Globals

    """
    # third party
    import sympc

    # syft relative
    from . import rst_share  # noqa: 401
    from . import session  # noqa: 401
    from . import share  # noqa: 401

    ast = Globals(client=client)

    modules: TypeList[TypeTuple[str, TypeAny]] = sympc.api.allowed_external_modules
    add_modules(ast, modules)

    classes: TypeList[TypeTuple[str, str, TypeAny]] = sympc.api.allowed_external_classes
    add_classes(ast, classes)

    attrs: TypeList[TypeTuple[str, str]] = sympc.api.allowed_external_attrs
    add_methods(ast, attrs)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)



# stdlib
import ast
from typing import Any
from typing import Callable
from typing import Dict
from typing import Union

# third party
from pandas import DataFrame

# syft relative
from ....core.io.connection import ClientConnection
from ....core.node.domain.client import DomainClient
from ...messages.dataset_messages import CreateDatasetMessage
from ...messages.dataset_messages import DeleteDatasetMessage
from ...messages.dataset_messages import GetDatasetMessage
from ...messages.dataset_messages import GetDatasetsMessage
from ...messages.dataset_messages import UpdateDatasetMessage
from ..enums import ResponseObjectEnum
from .request_api import GridRequestAPI


class DatasetRequestAPI(GridRequestAPI):
    def __init__(self, send: Callable, conn: ClientConnection, client: DomainClient):
        super().__init__(
            create_msg=CreateDatasetMessage,
            get_msg=GetDatasetMessage,
            get_all_msg=GetDatasetsMessage,
            update_msg=UpdateDatasetMessage,
            delete_msg=DeleteDatasetMessage,
            send=send,
            response_key=ResponseObjectEnum.DATASET,
        )
        self.conn = conn
        self.client = client

    def create(self, path: str) -> Dict[str, str]:  # type: ignore
        response = self.conn.send_files(path)  # type: ignore
        return self.to_obj(response)

    def __getitem__(self, key: Union[str, int, slice]) -> Any:
        # optionally we should be able to pass in the index of the dataset we want
        # according to the order displayed when displayed as a pandas table
        if isinstance(key, int):
            key = self.all()[key]["id"]

        return self.get(dataset_id=key)

    def __delitem__(self, key: str) -> Any:
        self.delete(dataset_id=key)

    def to_obj(self, result: Any) -> Any:
        dataset_obj = super().to_obj(result)
        dataset_obj.pandas = DataFrame(dataset_obj.data)
        datasets = []

        pointers = self.client.store
        for data in dataset_obj.data:
            _class_name = ResponseObjectEnum.DATA.capitalize()
            data_obj = type(_class_name, (object,), data)()
            data_obj.shape = ast.literal_eval(data_obj.shape)
            data_obj.pointer = pointers[data_obj.id.replace("-", "")]
            datasets.append(data_obj)

        dataset_obj.files = datasets
        type(dataset_obj).__getitem__ = lambda x, i: x.data[i]
        dataset_obj.client = self.client
        return Dataset(dataset_obj)

    def _repr_html_(self) -> str:
        """Displays a nice table when the object is returned in Jupyter Notebook"""
        return self.all(pandas=True)._repr_html_()  # type: ignore


class Dataset:
    def __init__(self, dataset_metadata: Any) -> None:
        self.dataset_metadata = dataset_metadata
        self.id = self.dataset_metadata.id
        self.tags = self.dataset_metadata.tags
        self.manifest = self.dataset_metadata.manifest
        self.pandas = self.dataset_metadata.pandas

    def __getitem__(self, key: Union[str, int]) -> Any:
        if isinstance(key, int):
            obj_id = self.dataset_metadata.data[key]["id"].replace("-", "")
            return self.dataset_metadata.client.store[obj_id]
        elif isinstance(key, str):
            return self.dataset_metadata.client.store[key.replace("-", "")]

    def _repr_html_(self) -> str:
        id = "<b>Id: </b>" + str(self.id) + "<br />"
        tags = "<b>Tags: </b>" + str(self.tags) + "<br />"
        manifest = "<b>Manifest: </b>" + str(self.manifest) + "<br /><br />"
        table = self.pandas._repr_html_()

        return self.__repr__() + "<br /><br />" + id + tags + manifest + table



# stdlib
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

# third party
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
import names
from pandas import DataFrame

# syft relative
from ...core.common.message import EventualSyftMessageWithoutReply
from ...core.common.message import ImmediateSyftMessageWithReply
from ...core.common.message import ImmediateSyftMessageWithoutReply
from ...core.common.message import SignedImmediateSyftMessageWithReply
from ...core.common.message import SignedImmediateSyftMessageWithoutReply
from ...core.common.message import SyftMessage
from ...core.common.serde.serialize import _serialize as serialize  # noqa: F401
from ...core.io.address import Address
from ...core.io.connection import ClientConnection
from ...core.io.location.specific import SpecificLocation
from ...core.io.route import SoloRoute
from ...core.node.common.client import Client
from ...core.node.device.client import DeviceClient
from ...core.node.domain.client import DomainClient
from ...core.node.network.client import NetworkClient
from ...core.node.vm.client import VirtualMachineClient
from ...core.pointer.pointer import Pointer
from ..messages.network_search_message import NetworkSearchMessage
from ..messages.setup_messages import GetSetUpMessage
from ..messages.transfer_messages import LoadObjectMessage
from .enums import PyGridClientEnums
from .enums import RequestAPIFields
from .exceptions import RequestAPIException
from .grid_connection import GridHTTPConnection
from .request_api.association_api import AssociationRequestAPI
from .request_api.dataset_api import DatasetRequestAPI
from .request_api.group_api import GroupRequestAPI
from .request_api.role_api import RoleRequestAPI
from .request_api.user_api import UserRequestAPI
from .request_api.worker_api import WorkerRequestAPI

DEFAULT_PYGRID_PORT = 5000
DEFAULT_PYGRID_ADDRESS = f"http://127.0.0.1:{DEFAULT_PYGRID_PORT}"
INSECURE_PYGRID_TOKEN = "9G9MJ06OQH"  # nosec


class GridClient(DomainClient):
    def __init__(
        self,
        credentials: Dict,
        url: str,
        conn_type: Type[ClientConnection],
        client_type: Type[Client],
        user_key: Optional[SigningKey] = None,
    ) -> None:

        # Use Server metadata
        # to build client route
        self.conn = conn_type(url=url)  # type: ignore
        self.client_type = client_type

        if credentials:
            metadata, _user_key = self.conn.login(credentials=credentials)  # type: ignore
            _user_key = SigningKey(_user_key.encode(), encoder=HexEncoder)
        else:
            metadata = self.conn._get_metadata()  # type: ignore
            if not user_key:
                _user_key = SigningKey.generate()
            else:
                _user_key = user_key

        (
            spec_location,
            name,
            client_id,
        ) = self.client_type.deserialize_client_metadata_from_node(metadata=metadata)

        # Create a new Solo Route using the selected connection type
        route = SoloRoute(destination=spec_location, connection=self.conn)

        location_args = self.__route_client_location(
            client_type=self.client_type, location=spec_location
        )

        self.proxy_address: Optional[Address] = None

        # Create a new client using the selected client type
        super().__init__(
            network=location_args[NetworkClient],
            domain=location_args[DomainClient],
            device=location_args[DeviceClient],
            vm=location_args[VirtualMachineClient],
            name=name,
            routes=[route],
            signing_key=_user_key,
        )

        self.groups = GroupRequestAPI(send=self.__perform_grid_request)
        self.users = UserRequestAPI(send=self.__perform_grid_request)
        self.roles = RoleRequestAPI(send=self.__perform_grid_request)
        self.workers = WorkerRequestAPI(
            send=self.__perform_grid_request, domain_client=self
        )
        self.association_requests = AssociationRequestAPI(
            send=self.__perform_grid_request
        )
        self.datasets = DatasetRequestAPI(
            send=self.__perform_grid_request, conn=self.conn, client=self
        )

    def load(
        self, obj_ptr: Type[Pointer], address: Address, pointable: bool = False
    ) -> None:
        content = {
            RequestAPIFields.ADDRESS: serialize(address)
            .SerializeToString()  # type: ignore
            .decode(PyGridClientEnums.ENCODING),
            RequestAPIFields.UID: str(obj_ptr.id_at_location.value),
            RequestAPIFields.POINTABLE: pointable,
        }
        self.__perform_grid_request(grid_msg=LoadObjectMessage, content=content)

    def setup(self, *, domain_name: Optional[str], **kwargs: Any) -> Any:

        if domain_name is None:
            domain_name = names.get_full_name() + "'s Domain"
            logging.info(
                "No Domain Name provided... picking randomly as: " + domain_name
            )

        if "token" not in kwargs.keys():
            kwargs["token"] = INSECURE_PYGRID_TOKEN
            logging.info(
                "No token provided... trying PyGrid Default (THIS IS NOT SAFE!!!): "
                + kwargs["token"]
            )

        kwargs["domain_name"] = domain_name

        response = self.conn.setup(**kwargs)  # type: ignore
        logging.info(response[RequestAPIFields.MESSAGE])

    def get_setup(self, **kwargs: Any) -> Any:
        return self.__perform_grid_request(grid_msg=GetSetUpMessage, content=kwargs)

    def search(self, query: List, pandas: bool = False) -> Any:
        response = self.__perform_grid_request(
            grid_msg=NetworkSearchMessage, content={RequestAPIFields.QUERY: query}
        )
        if pandas:
            response = DataFrame(response)

        return response

    def send_immediate_msg_with_reply(
        self,
        msg: Union[SignedImmediateSyftMessageWithReply, ImmediateSyftMessageWithReply],
        route_index: int = 0,
    ) -> Union[
        SignedImmediateSyftMessageWithoutReply, ImmediateSyftMessageWithoutReply
    ]:
        return super(GridClient, self).send_immediate_msg_with_reply(  # type: ignore
            msg=msg, route_index=route_index
        )

    def send_immediate_msg_without_reply(
        self,
        msg: Union[
            SignedImmediateSyftMessageWithoutReply, ImmediateSyftMessageWithoutReply
        ],
        route_index: int = 0,
    ) -> None:
        super(GridClient, self).send_immediate_msg_without_reply(
            msg=msg, route_index=route_index
        )

    def send_eventual_msg_without_reply(
        self,
        msg: EventualSyftMessageWithoutReply,
        route_index: int = 0,
    ) -> None:
        return super(GridClient, self).send_eventual_msg_without_reply(
            msg=msg, route_index=route_index
        )

    def __route_client_location(
        self, client_type: Any, location: SpecificLocation
    ) -> Dict[Any, Any]:
        locations: Dict[Any, Optional[SpecificLocation]] = {
            NetworkClient: None,
            DomainClient: None,
            DeviceClient: None,
            VirtualMachineClient: None,
        }
        locations[client_type] = location
        return locations

    def __perform_grid_request(
        self, grid_msg: Any, content: Optional[Dict[Any, Any]] = None
    ) -> Dict[Any, Any]:
        if content is None:
            content = {}
        signed_msg = self.__build_msg(grid_msg=grid_msg, content=content)
        response = self.send_immediate_msg_with_reply(msg=signed_msg)
        # print("got this intermediary", response, type(response))
        return self.__process_response(response=response)

    def __build_msg(self, grid_msg: Any, content: Dict[Any, Any]) -> Any:
        args = {
            RequestAPIFields.ADDRESS: self.address,
            RequestAPIFields.CONTENT: content,
            RequestAPIFields.REPLY_TO: self.address,
        }
        return grid_msg(**args).sign(signing_key=self.signing_key)

    def __process_response(self, response: SyftMessage) -> Dict[Any, Any]:
        # print("calling __process_response", response, type(response))
        if response.status_code == 200:  # type: ignore
            return response.content  # type: ignore
        else:
            raise RequestAPIException(message=response.content[RequestAPIFields.ERROR])  # type: ignore


def connect(
    url: str = DEFAULT_PYGRID_ADDRESS,
    conn_type: Type[ClientConnection] = GridHTTPConnection,
    credentials: Dict = {},
    user_key: Optional[SigningKey] = None,
) -> GridClient:
    return GridClient(
        credentials=credentials,
        url=url,
        conn_type=conn_type,
        client_type=DomainClient,
        user_key=user_key,
    )


def login(
    email: Optional[str] = None,
    password: Optional[str] = None,
    url: str = DEFAULT_PYGRID_ADDRESS,
    conn_type: Type[ClientConnection] = GridHTTPConnection,
) -> GridClient:

    if email is None or password is None:
        credentials = {}
        logging.info(
            "\n\nNo email and password defined in login() - connecting as anonymous user!!!\n"
        )
    else:
        credentials = {"email": email, "password": password}

    return connect(url=url, credentials=credentials, conn_type=conn_type)



# third party
from tokenizers import Tokenizer
from transformers import PreTrainedTokenizerFast

# syft relative
from ... import deserialize
from ...generate_wrapper import GenerateWrapper
from ...lib.python.primitive_factory import PrimitiveFactory
from ...lib.python.util import upcast
from ...proto.lib.transformers.tokenizerfast_pb2 import (
    TokenizerFast as TokenizerFast_PB,
)
from ..util import full_name_with_qualname


def object2proto(obj: PreTrainedTokenizerFast) -> TokenizerFast_PB:
    tokenizer_type = full_name_with_qualname(klass=type(obj))
    tokenizer_str = obj._tokenizer.to_str()

    kwargs = obj.special_tokens_map
    kwargs["name_or_path"] = obj.name_or_path
    kwargs["padding_side"] = obj.padding_side
    kwargs["model_max_length"] = obj.model_max_length
    kwargs = PrimitiveFactory.generate_primitive(value=kwargs)

    protobuf_tokenizer = TokenizerFast_PB(
        id=kwargs.id._object2proto(),
        tokenizer_type=tokenizer_type,
        tokenizer=tokenizer_str,
        kwargs=kwargs._object2proto(),
    )
    return protobuf_tokenizer


def proto2object(proto: TokenizerFast_PB) -> PreTrainedTokenizerFast:
    # TODO some subclasses of pretrainedtokenizerfast have required args,
    # cast every tokenizer as baseclass pretrainedtokenizerfast for now.
    # See  BertTokenizerFast.

    tokenizer_type = PreTrainedTokenizerFast

    _tokenizer = Tokenizer.from_str(proto.tokenizer)
    kwargs = deserialize(proto.kwargs)
    kwargs = upcast(kwargs)

    tokenizer = tokenizer_type(tokenizer_object=_tokenizer, **kwargs)
    return tokenizer


GenerateWrapper(
    wrapped_type=PreTrainedTokenizerFast,
    import_path="transformers.PreTrainedTokenizerFast",
    protobuf_scheme=TokenizerFast_PB,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
import io
import json
import sys
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

# third party
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from requests_toolbelt.multipart import encoder

# syft relative
from ...core.common.message import SyftMessage
from ...core.common.serde.serialize import _serialize
from ...proto.core.node.common.metadata_pb2 import Metadata as Metadata_PB
from ..client.enums import RequestAPIFields
from ..client.exceptions import RequestAPIException
from ..connections.http_connection import HTTPConnection


class GridHTTPConnection(HTTPConnection):
    LOGIN_ROUTE = "/users/login"
    SYFT_ROUTE = "/pysyft"
    SYFT_MULTIPART_ROUTE = "/pysyft_multipart"
    SIZE_THRESHOLD = 20971520  # 20 MB

    def __init__(self, url: str) -> None:
        self.base_url = url
        self.session_token: Optional[Dict[str, str]] = None

    def _send_msg(self, msg: SyftMessage) -> requests.Response:
        """
        Serializes Syft messages in json format and send it using HTTP protocol.
        NOTE: Auxiliary method to avoid code duplication and modularity.
        :return: returns requests.Response object containing a JSON serialized
        SyftMessage
        :rtype: requests.Response
        """

        header = {}

        if self.session_token:
            header["token"] = self.session_token

        header["Content-Type"] = "application/octet-stream"  # type: ignore

        # Perform HTTP request using base_url as a root address
        msg_bytes: bytes = _serialize(obj=msg, to_bytes=True)  # type: ignore

        if sys.getsizeof(msg_bytes) < GridHTTPConnection.SIZE_THRESHOLD:
            r = requests.post(
                url=self.base_url + GridHTTPConnection.SYFT_ROUTE,
                data=msg_bytes,
                headers=header,
            )
        else:
            r = self.send_streamed_messages(blob_message=msg_bytes)

        # Return request's response object
        # r.text provides the response body as a str
        return r

    def login(self, credentials: Dict) -> Tuple:
        # Login request
        response = requests.post(
            url=self.base_url + GridHTTPConnection.LOGIN_ROUTE, json=credentials
        )

        # Response
        content = json.loads(response.text)

        # If fail
        if response.status_code != requests.codes.ok:
            raise Exception(content["error"])

        metadata = content["metadata"].encode("ISO-8859-1")
        metadata_pb = Metadata_PB()
        metadata_pb.ParseFromString(metadata)

        # If success
        # Save session token
        self.session_token = content["token"]

        # Return node metadata / user private key
        return (metadata_pb, content["key"])

    def _get_metadata(self) -> Tuple:
        """Request Node's metadata
        :return: returns node metadata
        :rtype: str of bytes
        """

        # allow retry when connecting in CI
        session = requests.Session()
        retry = Retry(connect=3, backoff_factor=0.5)
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        response = session.get(self.base_url + "/metadata")
        content = json.loads(response.text)

        metadata = content["metadata"].encode("ISO-8859-1")
        metadata_pb = Metadata_PB()
        metadata_pb.ParseFromString(metadata)

        return metadata_pb

    def setup(self, **content: Dict[str, Any]) -> Any:
        response = json.loads(
            requests.post(self.base_url + "/setup", json=content).text
        )
        if response.get(RequestAPIFields.MESSAGE, None):
            return response
        else:
            raise RequestAPIException(response.get(RequestAPIFields.ERROR))

    def send_files(self, file_path: str) -> Dict[str, Any]:
        session = requests.Session()

        with open(file_path, "rb") as f:

            form = encoder.MultipartEncoder(
                {
                    "file": (file_path, f, "application/octet-stream"),
                }
            )

            headers = {
                "Prefer": "respond-async",
                "Content-Type": form.content_type,
                "token": self.session_token,
            }

            resp = session.post(
                self.base_url + "/data-centric/datasets", headers=headers, data=form
            )

        session.close()

        return json.loads(resp.content)

    def send_streamed_messages(self, blob_message: bytes) -> requests.Response:
        session = requests.Session()
        with io.BytesIO(blob_message) as msg:
            form = encoder.MultipartEncoder(
                {
                    "file": ("message", msg.read(), "application/octet-stream"),
                }
            )

            headers = {
                "Prefer": "respond-async",
                "Content-Type": form.content_type,
            }

            resp = session.post(
                self.base_url + GridHTTPConnection.SYFT_MULTIPART_ROUTE,
                headers=headers,
                data=form,
            )

        session.close()
        return resp



# third party
from statsmodels.genmod.generalized_linear_model import GLMResultsWrapper

# syft relative
from ...generate_wrapper import GenerateWrapper
from ...lib.python.primitive_factory import PrimitiveFactory
from ...lib.python.string import String
from ...proto.lib.statsmodels.results_pb2 import ResultsProto


def object2proto(obj: GLMResultsWrapper) -> ResultsProto:
    summary = obj.summary().as_csv()
    summary_prim = PrimitiveFactory.generate_primitive(value=summary)
    summary_proto = summary_prim._object2proto()
    return ResultsProto(summary=summary_proto)


def proto2object(proto: ResultsProto) -> str:
    return str(String._proto2object(proto.summary))


GenerateWrapper(
    wrapped_type=GLMResultsWrapper,
    import_path="statsmodels.genmod.generalized_linear_model.GLMResultsWrapper",
    protobuf_scheme=ResultsProto,
    type_object2proto=object2proto,
    type_proto2object=proto2object,
)



# stdlib
from typing import Any
from typing import Callable
from typing import Dict

# syft relative
from ...messages.association_messages import DeleteAssociationRequestMessage
from ...messages.association_messages import GetAssociationRequestMessage
from ...messages.association_messages import GetAssociationRequestsMessage
from ...messages.association_messages import RespondAssociationRequestMessage
from ...messages.association_messages import SendAssociationRequestMessage
from ..enums import AssociationRequestResponses
from ..enums import RequestAPIFields
from ..enums import ResponseObjectEnum
from ..exceptions import PyGridClientException
from .request_api import GridRequestAPI


class AssociationRequestAPI(GridRequestAPI):
    def __init__(self, send: Callable):
        super().__init__(
            create_msg=SendAssociationRequestMessage,
            get_msg=GetAssociationRequestMessage,
            get_all_msg=GetAssociationRequestsMessage,
            delete_msg=DeleteAssociationRequestMessage,
            send=send,
            response_key=ResponseObjectEnum.ASSOCIATION_REQUEST,
        )

    def update(self, **kwargs: Any) -> Dict[Any, Any]:  # type: ignore
        raise PyGridClientException(
            "You can not update an association request, try to send another one instead."
        )

    def __getitem__(self, key: int) -> Any:
        return self.get(association_request_id=key)

    def __delitem__(self, key: int) -> None:
        self.delete(association_request_id=key)

    def to_obj(self, result: Dict[Any, Any]) -> Any:
        _association_obj = super().to_obj(result)

        _content = {
            RequestAPIFields.ADDRESS: _association_obj.address,
            RequestAPIFields.HANDSHAKE: _association_obj.handshake_value,
            RequestAPIFields.SENDER_ADDRESS: _association_obj.sender_address,
        }

        def _accept() -> Dict[str, str]:
            _content[RequestAPIFields.VALUE] = AssociationRequestResponses.ACCEPT
            return self.send(
                grid_msg=RespondAssociationRequestMessage, content=_content
            )

        def _deny() -> Dict[str, str]:
            _content[RequestAPIFields.VALUE] = AssociationRequestResponses.DENY
            return self.send(
                grid_msg=RespondAssociationRequestMessage, content=_content
            )

        _association_obj.accept = _accept
        _association_obj.deny = _deny

        return _association_obj



# stdlib
import functools
from typing import Any as TypeAny
from typing import List as TypeList
from typing import Tuple as TypeTuple

# third party
import gym

# syft relative
from ...ast import add_classes
from ...ast import add_methods
from ...ast import add_modules
from ...ast.globals import Globals
from ..util import generic_update_ast

LIB_NAME = "gym"
PACKAGE_SUPPORT = {
    "lib": LIB_NAME,
}


def create_ast(client: TypeAny = None) -> Globals:
    ast = Globals(client)

    modules: TypeList[TypeTuple[str, TypeAny]] = [
        ("gym", gym),
        ("gym.wrappers", gym.wrappers),
        ("gym.wrappers.time_limit", gym.wrappers.time_limit),
    ]
    classes: TypeList[TypeTuple[str, str, TypeAny]] = [
        (
            "gym.wrappers.time_limit.TimeLimit",
            "gym.wrappers.time_limit.TimeLimit",
            gym.wrappers.time_limit.TimeLimit,
        ),
        (
            "gym.Wrapper",
            "gym.Wrapper",
            gym.Wrapper,
        ),
    ]

    methods = [
        ("gym.make", "gym.wrappers.time_limit.TimeLimit"),
        ("gym.wrappers.time_limit.TimeLimit.seed", "syft.lib.python.List"),
        ("gym.wrappers.time_limit.TimeLimit.reset", "numpy.ndarray"),
        ("gym.wrappers.time_limit.TimeLimit.step", "syft.lib.python.Tuple"),
    ]

    add_modules(ast, modules)
    add_classes(ast, classes)
    add_methods(ast, methods)

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast


update_ast = functools.partial(generic_update_ast, LIB_NAME, create_ast)
